"""_gcs_functions.py: Functions for working with grid connection sets."""

version = '20th October 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.fault as rqf
import resqpy.grid as grr
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
import resqpy.property as rqp


def pinchout_connection_set(grid, skip_inactive = True, feature_name = 'pinchout'):
    """Returns a new GridConnectionSet representing non-standard K face connections across pinchouts.

    arguments:
       grid (grid.Grid): the grid for which a pinchout connection set is required
       skip_inactive (boolean, default True): if True, connections are not included where there is an inactive cell
          above or below the pinchout; if False, such connections are included
       feature_name (string, default 'pinchout'): the name to use as citation title in the feature and interpretation

    notes:
       this function does not write to hdf5, nor create xml for the new grid connection set;
       however, it does create one feature and a corresponding interpretation and creates xml for those
    """

    assert grid is not None

    po = grid.pinched_out()
    dead = grid.extract_inactive_mask() if skip_inactive else None

    cip_list = []  # cell index pair list

    for j in range(grid.nj):
        for i in range(grid.ni):
            ka = 0
            while True:
                while ka < grid.nk - 1 and po[ka, j, i]:
                    ka += 1
                while ka < grid.nk - 1 and not po[ka + 1, j, i]:
                    ka += 1
                if ka >= grid.nk - 1:
                    break
                # ka now in non-pinched out cell above pinchout
                if (skip_inactive and dead[ka, j, i]) or (grid.k_gaps and grid.k_gap_after_array[ka]):
                    ka += 1
                    continue
                kb = ka + 1
                while kb < grid.nk and po[kb, j, i]:
                    kb += 1
                if kb >= grid.nk:
                    break
                if skip_inactive and dead[kb, j, i]:
                    ka = kb + 1
                    continue
                # kb now beneath pinchout
                cip_list.append((grid.natural_cell_index((ka, j, i)), grid.natural_cell_index((kb, j, i))))
                ka = kb + 1

    log.debug(f'{len(cip_list)} pinchout connections found')

    pcs = _make_k_gcs_from_cip_list(grid, cip_list, feature_name)

    return pcs


def k_gap_connection_set(grid, skip_inactive = True, feature_name = 'k gap connection', tolerance = 0.001):
    """Returns a new GridConnectionSet representing K face connections where a K gap is zero thickness.

    arguments:
       grid (grid.Grid): the grid for which a K gap connection set is required
       skip_inactive (boolean, default True): if True, connections are not included where there is an inactive cell
          above or below the pinchout; if False, such connections are included
       feature_name (string, default 'pinchout'): the name to use as citation title in the feature and interpretation
       tolerance (float, default 0.001): the minimum vertical distance below which a K gap is deemed to be zero
          thickness; units are implicitly the z units of the coordinate reference system used by grid

    notes:
       this function does not write to hdf5, nor create xml for the new grid connection set;
       however, it does create one feature and a corresponding interpretation and creates xml for those;
       note that the entries in the connection set will be for logically K-neighbouring pairs of cells – such pairs
       are omitted from the standard transmissibilities due to the presence of the K gap layer
    """

    assert grid is not None
    if not grid.k_gaps:
        return None

    p = grid.points_ref(masked = False)
    dead = grid.extract_inactive_mask() if skip_inactive else None

    grid.set_crs()
    flip_z = (grid.k_direction_is_down != grid.crs.z_inc_down)

    cip_list = []  # cell index pair list

    for k in range(grid.nk - 1):
        if not grid.k_gap_after_array[k]:
            continue
        k_gap_pillar_z = p[grid.k_raw_index_array[k + 1]][..., 2] - p[grid.k_raw_index_array[k] + 1][..., 2]
        if grid.has_split_coordinate_lines:
            pfc = grid.create_column_pillar_mapping()  # pillars for column
            k_gap_z = 0.25 * np.sum(k_gap_pillar_z[pfc], axis = (2, 3))  # resulting shape (nj, ni)
        else:
            k_gap_z = 0.25 * (k_gap_pillar_z[:-1, :-1] + k_gap_pillar_z[:-1, 1:] + k_gap_pillar_z[1:, :-1] +
                              k_gap_pillar_z[1:, 1:])  # shape (nj, ni)
        if flip_z:
            k_gap_z = -k_gap_z
        layer_mask = np.logical_and(np.logical_not(np.isnan(k_gap_z)), k_gap_z < tolerance)
        if skip_inactive:
            layer_mask = np.logical_and(layer_mask, np.logical_not(np.logical_or(dead[k], dead[k + 1])))
        # layer mask now boolean array of shape (nj, ni) set True where connection needed
        ji_list = np.stack(np.where(layer_mask)).T  # numpy array being list of [j, i] pairs
        for (j, i) in ji_list:
            cip_list.append((grid.natural_cell_index((k, j, i)), grid.natural_cell_index((k + 1, j, i))))

    log.debug(f'{len(cip_list)} k gap connections found')

    kgcs = _make_k_gcs_from_cip_list(grid, cip_list, feature_name)

    return kgcs


def cell_set_skin_connection_set(grid,
                                 cell_set_mask,
                                 feature_name,
                                 feature_type = 'geobody boundary',
                                 title = None,
                                 create_organizing_objects_where_needed = True):
    """Add a grid connection set containing external faces of selected set of cells.

    arguments:
        grid (Grid): the grid for which the connection set is required
        cell_set_mask (numpy bool array of shape grid.extent_kji): True values identify cells included in the set
        feature_name (str): the name of the skin feature
        feature_type (str, default 'geobody boundary'): 'fault', 'horizon' or 'geobody boundary'
        title (str, optional): the citation title to use for the gcs; defaults to the feature_name
        create_organizing_objects_where_needed (bool, default True): if True, feature and interpretation
            objects will be created if they do not exist

    returns:
        the newly created grid connection set

    notes:
        this function does not take into consideration split pillars, it assumes cells are neighbouring based
        on the cell indices; faces on the outer skin of the grid are not included in the connection set;
        any cell face between a cell in the cell set and one not in it will be included in the connection set,
        therefore the set may contain internal skin faces as well as the outer skin
    """

    from resqpy.fault._grid_connection_set import GridConnectionSet

    assert grid is not None
    assert cell_set_mask.shape == tuple(grid.extent_kji)
    assert feature_name, 'no feature name given for cell set skin connection set'
    assert feature_type in ['geobody boundary', 'fault', 'horizon'], f'invalid feature type: {feature_type}'
    if not title:
        title = feature_name

    k_faces = np.zeros((grid.nk - 1, grid.nj, grid.ni), dtype = bool)
    j_faces = np.zeros((grid.nk, grid.nj - 1, grid.ni), dtype = bool)
    i_faces = np.zeros((grid.nk, grid.nj, grid.ni - 1), dtype = bool)

    if grid.nk > 1:
        k_faces[np.where(cell_set_mask[1:, :, :] != cell_set_mask[:-1, :, :])] = True
    if grid.nj > 1:
        j_faces[np.where(cell_set_mask[:, 1:, :] != cell_set_mask[:, :-1, :])] = True
    if grid.ni > 1:
        i_faces[np.where(cell_set_mask[:, :, 1:] != cell_set_mask[:, :, :-1])] = True

    gcs = GridConnectionSet(grid.model,
                            grid = grid,
                            k_faces = k_faces,
                            j_faces = j_faces,
                            i_faces = i_faces,
                            feature_name = feature_name,
                            feature_type = feature_type,
                            create_organizing_objects_where_needed = create_organizing_objects_where_needed,
                            title = title)

    return gcs


def add_connection_set_and_tmults(model, fault_incl, tmult_dict = None):
    """Add a grid connection set to a resqml model, based on a fault include file and a dictionary of fault:tmult pairs.

    Grid connection set added to resqml model, with extra_metadata on the fault interpretation containing the MULTFL values

    arguments:
       model: resqml model object
       fault_incl: fullpath to fault include file or list of fullpaths to fault include files
       tmult_dict: dictionary of fault name/transmissibility multiplier pairs (must align with faults in include file).
          Optional, if blank values in the fault.include file will be used instead
    returns:
       grid connection set uuid
    """

    from resqpy.fault._grid_connection_set import GridConnectionSet

    if isinstance(fault_incl, list):
        if len(fault_incl) > 1:
            # Making a concatenated version of the faultincl files
            # TODO: perhaps a better/more unique name and location could be used in future?
            temp_faults = os.path.join(os.path.dirname(model.epc_file), 'faults_combined_temp.txt')
            with open(temp_faults, 'w') as outfile:
                log.debug("combining multiple include files into one")
                for fname in fault_incl:
                    with open(fname) as infile:
                        for line in infile:
                            outfile.write(line.rstrip() + '\n')
        else:
            temp_faults = fault_incl[0]
    else:
        temp_faults = fault_incl

    log.info("Creating grid connection set")
    gcs = GridConnectionSet(parent_model = model,
                            ascii_load_format = 'nexus',
                            ascii_file = temp_faults,
                            create_organizing_objects_where_needed = True,
                            create_transmissibility_multiplier_property = True,
                            fault_tmult_dict = tmult_dict)
    gcs.write_hdf5(model.h5_file_name())
    gcs.create_xml(model.h5_uuid())

    model.store_epc()

    # clean up
    temp_combined_file = os.path.join(os.path.dirname(model.epc_file), 'faults_combined_temp.txt')
    if os.path.exists(temp_combined_file):
        os.remove(temp_combined_file)

    log.info("Grid connection set added")

    return gcs.uuid


def grid_columns_property_from_gcs_property(model,
                                            gcs_property_uuid,
                                            null_value = np.nan,
                                            title = None,
                                            multiple_handling = 'default'):
    """Derives a new grid columns property (map) from a single-grid gcs property using values for k faces.

    arguments:
        model (Model): the model in which the existing objects are to be found and the new property added
        gcs_property_uuid (UUID): the uuid of the existing grid connection set property
        null_value (float or int, default NaN): the value to use in columns where no K faces are present in the gcs
        title (str, optional): the title for the new grid property; defaults to that of the gcs property
        multiple_handling (str, default 'mean'): one of 'default', 'mean', 'min', 'max', 'min_k', 'max_k', 'exception';
            determines how a value is generated when more than one K face is present in the gcs for a column

    returns:
        uuid of the newly created Property (RESQML ContinuousProperty, DiscreteProperty, CategoricalProperty or
        PointsProperty)

    notes:
        the grid connection set which is the support for gcs_property must involve only one grid;
        the resulting columns grid property is of the same class as the original gcs property;
        the write_hdf() and create_xml() methods are called by this function, for the new property,
        which is added to the model;
        the default multiple handling mode is mean for continuous data, any for discrete (inc categorical);
        in the case of discrete (including categorical) data, a null_value of NaN will be changed to -1
    """
    assert multiple_handling in ['default', 'mean', 'min', 'max', 'any', 'exception']
    assert gcs_property_uuid is not None and bu.is_uuid(gcs_property_uuid)
    gcs_property = rqp.Property(model, uuid = gcs_property_uuid)
    assert gcs_property is not None
    support_uuid = gcs_property.collection.support_uuid
    assert support_uuid is not None
    assert model.type_of_uuid(support_uuid, strip_obj = True) == 'GridConnectionSetRepresentation'
    gcs = rqf.GridConnectionSet(model, uuid = support_uuid)
    gcs_prop_array = gcs_property.array_ref()
    if gcs_property.is_continuous():
        dtype = float
        if multiple_handling == 'default':
            multiple_handling = 'mean'
    else:
        dtype = int
        if null_value == np.nan:
            null_value = -1
        elif type(null_value) is float:
            null_value = int(null_value)
        if multiple_handling == 'default':
            multiple_handling = 'any'
        assert multiple_handling != 'mean', 'mean specified as multiple handling for non-continuous property'
    assert gcs.number_of_grids() == 1, 'only single grid gcs supported for grid columns property derivation'
    grid = gcs.grid_list[0]
    if gcs_property.is_points():
        map_shape = (grid.nj, grid.ni, 3)
    else:
        map_shape = (grid.nj, grid.ni)
    assert gcs_property.count() == 1
    map = np.full(map_shape, null_value, dtype = dtype)
    count_per_col = np.zeros((grid.nj, grid.ni), dtype = int)
    cells_and_faces = gcs.list_of_cell_face_pairs_for_feature_index(None)
    assert cells_and_faces is not None and len(cells_and_faces) == 2
    cell_pairs, face_pairs = cells_and_faces
    if cell_pairs is None or len(cell_pairs) == 0:
        log.warning('no faces found for grid connection set property {gcs_property.title}')
        return None
    assert len(cell_pairs) == len(face_pairs)
    assert len(cell_pairs) == len(gcs_prop_array)
    for index in range(len(cell_pairs)):
        if face_pairs[index, 0, 0] != 0 or face_pairs[index, 1, 0] != 0:
            continue  # not a K face
        col_j, col_i = cell_pairs[index, 0, 1:]  # assume paired cell is in same column!
        if count_per_col[col_j, col_i] == 0 or multiple_handling == 'any':
            map[col_j, col_i] = gcs_prop_array[index]
        elif multiple_handling == 'mean':
            map[col_j, col_i] = (((map[col_j, col_i] * count_per_col[col_j, col_i]) + gcs_prop_array[index]) /
                                 float(count_per_col[col_j, col_i] + 1))
        elif multiple_handling == 'min':
            if gcs_prop_array[index] < map[col_j, col_i]:
                map[col_j, col_i] = gcs_prop_array[index]
        elif multiple_handling == 'max':
            if gcs_prop_array[index] > map[col_j, col_i]:
                map[col_j, col_i] = gcs_prop_array[index]
        else:
            raise ValueError('multiple grid connection set K faces found for column')
        count_per_col[col_j, col_i] += 1
    # create an empty PropertyCollection and add the map data as a new property
    time_index = gcs_property.time_index()
    pc = rqp.PropertyCollection()
    pc.set_support(support = grid)
    pc.add_cached_array_to_imported_list(
        map,
        source_info = f'{gcs.title} property {gcs_property.title} map',
        keyword = title if title else gcs_property.title,
        discrete = not gcs_property.is_continuous(),
        uom = gcs_property.uom(),
        time_index = time_index,
        null_value = None if gcs_property.is_continuous() else null_value,
        property_kind = gcs_property.property_kind(),
        local_property_kind_uuid = gcs_property.local_property_kind_uuid(),
        facet_type = gcs_property.facet_type(),
        facet = gcs_property.facet(),
        realization = None,  # do we want to preserve the realisation number?
        indexable_element = 'columns',
        points = gcs_property.is_points())
    time_series_uuid = pc.write_hdf5_for_imported_list()
    string_lookup_uuid = gcs_property.string_lookup_uuid()
    time_series_uuid = None if time_index is None else gcs_property.time_series_uuid()
    new_uuids = pc.create_xml_for_imported_list_and_add_parts_to_model(time_series_uuid = time_series_uuid,
                                                                       string_lookup_uuid = string_lookup_uuid)
    assert len(new_uuids) == 1
    return new_uuids[0]


def combined_tr_mult_from_gcs_mults(model,
                                    gcs_tr_mult_uuid_list,
                                    merge_mode = 'minimum',
                                    sided = None,
                                    fill_value = 1.0,
                                    active_only = True,
                                    apply_baffles = False):
    """Returns a triplet of transmissibility multiplier arrays over grid faces by combining those from gcs'es.

    arguments:
        model (Model): the model containing all the relevant objects
        gcs_tr_mult_uuid_list (list of UUID): uuids of the individual grid connection set transmissibility
            multiplier properties to be combined
        merge_mode (str, default 'minimum'): one of 'minimum', 'multiply', 'maximum', 'exception'; how to
            handle multiple values applicable to the same grid face
        sided (bool, optional): whether to apply values on both sides of each gcs cell-face pair; if None, will
            default to False if merge mode is multiply, True otherwise
        fill_value (float, optional): the value to use for grid faces not present in any of the gcs'es;
            if None, NaN will be used
        active_only (bool, default True): if True and an active property exists for a grid connection set,
            then only active faces are used when combining to make the grid face arrays
        apply_baffles (bool, default False): if True, where a baffle property exists for a grid connection
            set, a transmissibility multiplier of zero will be used for faces marked as True, overriding the
            multiplier property values at such faces

    returns:
        triple numpy float arrays being transmissibility multipliers for K, J, and I grid faces; arrays have
            shapes (nk + 1, nj, ni), (nk, nj + 1, ni), and (nk, nj, ni + 1) respectively

    notes:
        each gcs, which is the supporting representation for each input tr mult property, must be for a single
        grid and that grid must be the same for all the gcs'es
    """

    assert merge_mode in ['minimum', 'multiply', 'maximum', 'exception']
    assert gcs_tr_mult_uuid_list is not None and len(gcs_tr_mult_uuid_list) > 0
    if sided is None:
        sided = (merge_mode != 'multiply')
    elif sided and merge_mode == 'multiply':
        log.error('using a gcs transmissibility multiplier merge mode of multiply is not compatible with sided True')
        # carry on anyway!

    grid = None

    for tr_mult_uuid in gcs_tr_mult_uuid_list:

        tr_mult = rqp.Property(model, uuid = tr_mult_uuid)
        assert tr_mult is not None
        gcs_uuid = tr_mult.collection.support_uuid
        assert gcs_uuid is not None
        gcs = rqf.GridConnectionSet(model, uuid = gcs_uuid)
        assert gcs is not None
        assert gcs.number_of_grids() == 1

        baffle_uuid = None
        if apply_baffles:
            baffle_part = rqp.property_part(model,
                                            obj_type = 'Discrete',
                                            property_kind = 'baffle',
                                            related_uuid = gcs.uuid)
            if baffle_part is not None:
                baffle_uuid = model.uuid_for_part(baffle_part)

        if grid is None:  # first gcs: grab grid and initialise combined tr mult arrays
            grid = gcs.grid_list[0]
            combo_trm_k = np.full((grid.nk + 1, grid.nj, grid.ni), np.NaN, dtype = float)
            combo_trm_j = np.full((grid.nk, grid.nj + 1, grid.ni), np.NaN, dtype = float)
            combo_trm_i = np.full((grid.nk, grid.nj, grid.ni + 1), np.NaN, dtype = float)
        else:  # check same grid is referenced by this gcs
            assert bu.matching_uuids(gcs.grid_list[0].uuid, grid.uuid)

        # get gcs tr mult data in form of triplet of grid faces arrays
        gcs_trm_k, gcs_trm_j, gcs_trm_i = gcs.grid_face_arrays(tr_mult_uuid,
                                                               default_value = np.NaN,
                                                               active_only = active_only,
                                                               lazy = not sided,
                                                               baffle_uuid = baffle_uuid)
        assert all([trm is not None for trm in (gcs_trm_k, gcs_trm_j, gcs_trm_i)])

        # merge in each of the three directional face arrays for this gcs with combined arrays
        for (combo_trm, gcs_trm) in [(combo_trm_k, gcs_trm_k), (combo_trm_j, gcs_trm_j), (combo_trm_i, gcs_trm_i)]:
            mask = np.logical_not(np.isnan(gcs_trm))  # true where this tr mult is present
            clash_mask = np.logical_and(mask, np.logical_not(np.isnan(combo_trm)))  # true where combined value clashes
            if np.any(clash_mask):
                if merge_mode == 'exception':
                    raise ValueError('gcs transmissibility multiplier conflict when merging')
                if merge_mode == 'minimum':
                    combo_trm[:] = np.where(clash_mask, np.minimum(combo_trm, gcs_trm), combo_trm)
                elif merge_mode == 'maximum':
                    combo_trm[:] = np.where(clash_mask, np.maximum(combo_trm, gcs_trm), combo_trm)
                elif merge_mode == 'multiply':
                    combo_trm[:] = np.where(clash_mask, combo_trm * gcs_trm, combo_trm)
                else:
                    raise Exception(f'code failure with unrecognised merge mode {merge_mode}')
                mask = np.logical_and(mask,
                                      np.logical_not(clash_mask))  # remove clash faces from mask (already handled)
            if np.any(mask):
                combo_trm[:] = np.where(mask, gcs_trm, combo_trm)  # update combined array from individual array

    # for each of the 3 combined tr mult arrays, replace unused values with the default fill value
    # also check that any set values are non-negative
    for combo_trm in (combo_trm_k, combo_trm_j, combo_trm_i):
        if fill_value is not None and not np.isnan(fill_value):
            combo_trm[:] = np.where(np.isnan(combo_trm), fill_value, combo_trm)
            assert np.all(combo_trm >= 0.0)
        else:
            assert np.all(np.logical_or(np.isnan(combo_trm), combo_trm >= 0.0))

    return (combo_trm_k, combo_trm_j, combo_trm_i)


# fault face table pandas dataframe functions
# these functions are for processing dataframes that have been read from (or to be written to) simulator ascii files


def zero_base_cell_indices_in_faces_df(faces, reverse = False):
    """Decrements all the cell indices in the fault face dataframe, in situ (or increments if reverse is True)."""

    if reverse:
        offset = 1
    else:
        offset = -1
    for col in ['i1', 'i2', 'j1', 'j2', 'k1', 'k2']:
        temp = faces[col] + offset
        faces[col] = temp


def standardize_face_indicator_in_faces_df(faces):
    """Sets face indicators to uppercase I, J or K, always with + or - following direction, in situ."""

    # todo: convert XYZ into IJK respectively?
    temp = faces['face'].copy().str.upper()
    temp[faces['face'].str.len() == 1] = faces['face'][faces['face'].str.len() == 1] + '+'
    for u in temp.unique():
        s = str(u)
        assert len(s) == 2, 'incorrect length to face string in fault dataframe: ' + s
        assert s[0] in 'IJK', 'unknown direction (axis) character in fault dataframe: ' + s
        assert s[1] in '+-', 'unknown face polarity character in fault dataframe: ' + s
    faces['face'] = temp


def remove_external_faces_from_faces_df(faces, extent_kji, remove_all_k_faces = False):
    """Returns a subset of the rows of faces dataframe, excluding rows on external faces."""

    # NB: assumes cell indices have been converted to zero based
    # NB: ignores grid column, ie. assumes extent_kji is applicable to all rows
    # NB: assumes single layer of cells is specified in the direction of the face
    filtered = []
    max_k0 = extent_kji[0] - 1
    max_j0 = extent_kji[1] - 1
    max_i0 = extent_kji[2] - 1
    for i in range(len(faces)):
        entry = faces.iloc[i]
        f = entry['face']
        if ((entry['i1'] <= 0 and f == 'I-') or (entry['j1'] <= 0 and f == 'J-') or (entry['k1'] <= 0 and f == 'K-') or
            (entry['i2'] >= max_i0 and f == 'I+') or (entry['j2'] >= max_j0 and f == 'J+') or
            (entry['k2'] >= max_k0 and f == 'K+')):
            continue
        if remove_all_k_faces and f[0] == 'K':
            continue
        filtered.append(i)
    return faces.loc[filtered]


def _make_k_gcs_from_cip_list(grid, cip_list, feature_name):
    # cip (cell index pair) list contains pairs of natural cell indices for which k connection is required
    # first of pair is layer above (lower k to be precise), second is below (higher k)
    # called by pinchout_connection_set() and k_gap_connection_set() functions

    from resqpy.fault._grid_connection_set import GridConnectionSet

    count = len(cip_list)

    if count == 0:
        return None

    pcs = GridConnectionSet(grid.model)
    pcs.grid_list = [grid]
    pcs.count = count
    pcs.grid_index_pairs = np.zeros((count, 2), dtype = int)
    pcs.cell_index_pairs = np.array(cip_list, dtype = int)
    pcs.face_index_pairs = np.zeros((count, 2), dtype = int)  # initialize to top faces
    pcs.face_index_pairs[:, 0] = 1  # bottom face of cells above pinchout

    pcs.feature_indices = np.zeros(count, dtype = int)  # could create seperate features by layer above or below?
    gbf = rqo.GeneticBoundaryFeature(grid.model, kind = 'horizon', feature_name = feature_name)
    gbf_root = gbf.create_xml()
    fi = rqo.HorizonInterpretation(grid.model, genetic_boundary_feature = gbf)
    fi_root = fi.create_xml(gbf_root, title_suffix = None)
    fi_uuid = rqet.uuid_for_part_root(fi_root)

    pcs.feature_list = [('obj_HorizonInterpretation', fi_uuid, str(feature_name))]

    return pcs


def _triangulate_unsplit_grid_connection_set(gcs, feature_index = None):
    # returns triangulation as indices into grid.points_cached, for a single grid gcs only

    def flat_point_index(grid, cell_kji0, corner_index):
        ci = cell_kji0 + corner_index
        return (ci[0] * (grid.nj + 1) + ci[1]) * (grid.ni + 1) + ci[2]

    def add_to_tri(grid, tri, index, cell_kji0, axis, polarity):
        kjip = np.zeros(3, dtype = 'int')
        kjip[axis] = polarity
        a = (axis + 1) % 3
        b = (a + 1) % 3
        tri[index, :, 0] = flat_point_index(grid, cell_kji0, kjip)
        kjip[a] = 1
        kjip[b] = 1
        tri[index, :, 1] = flat_point_index(grid, cell_kji0, kjip)
        kjip[a] = 0
        tri[index, 0, 2] = flat_point_index(grid, cell_kji0, kjip)
        kjip[a] = 1
        kjip[b] = 0
        tri[index, 1, 2] = flat_point_index(grid, cell_kji0, kjip)

    if feature_index is None:
        assert gcs.number_of_features() == 1, 'no gcs feature selected for triangulation'
        feature_index = 0

    assert gcs.number_of_grids() == 1, 'more than one grid references by gcs'
    grid = gcs.grid_list[0]
    if isinstance(grid, grr.RegularGrid):
        grid.make_regular_points_cached(apply_origin_offset = False)
    else:
        assert not grid.has_split_coordinate_lines, 'grid is faulted'
        grid.cache_all_geometry_arrays()

    feature_name = gcs.feature_name_for_feature_index(feature_index)
    ci_pairs, fi_pairs = gcs.list_of_cell_face_pairs_for_feature_index(feature_index)
    cell_index_pairs = ci_pairs.copy()
    face_index_pairs = fi_pairs.copy()
    if cell_index_pairs is None or face_index_pairs is None:
        log.warning(f'no cell face pairs found in grid connection set for feature: {feature_name}')
        return None
    connection_count = cell_index_pairs.shape[0]

    tri_extent = (connection_count, 2, 3)  # 2 triangles per face, will get flattened to (-1, 3)
    tri = np.zeros(tri_extent, dtype = 'int')  # point indices
    tri_index = 0

    for connection in range(connection_count):
        cell_pair = cell_index_pairs[connection]  # shape (2, 3); values kji0
        axis, polarity = face_index_pairs[connection, 0]
        add_to_tri(grid, tri, tri_index, cell_pair[0], axis, polarity)
        tri_index += 1

    result = tri[:tri_index, :, :].reshape((-1, 3))
    if result.size == 0:
        return None

    return result



"""A function for writing out a Nexus CORP file using supplied grid geometry"""

# Nexus is a trademark of Halliburton

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.grid_functions as gf
import resqpy.olio.xml_et as rqet
import resqpy.olio.trademark as tm
import resqpy.olio.write_data as wd
import resqpy.crs as rqc
import resqpy.weights_and_measures as wam


def write_nexus_corp(grid,
                     file_name,
                     local_coords = False,
                     global_xy_units = None,
                     global_z_units = None,
                     global_z_increasing_downward = True,
                     nexus_unit_system = None,
                     write_nx_ny_nz = False,
                     write_units_keyword = False,
                     write_rh_keyword_if_needed = False,
                     write_corp_keyword = False,
                     use_binary = False,
                     binary_only = False,
                     nan_substitute_value = None):
    """Write grid geometry to file in Nexus CORP ordering."""

    log.info('caching Nexus corner points')
    tm.log_nexus_tm('info')
    grid.corner_points(cache_cp_array = True)
    log.debug('duplicating Nexus corner points')
    cp = grid.array_corner_points.copy()
    log.debug('resequencing duplicated Nexus corner points')
    gf.resequence_nexus_corp(cp, eight_mode = False, undo = True)
    corp_extent = np.zeros(3, dtype = 'int')
    corp_extent[0] = grid.cell_count()  # total number of cells in grid
    corp_extent[1] = 8  # 8 corners of cell: k -/+; j -/+; i -/+
    corp_extent[2] = 3  # x, y, z
    ijk_right_handed = grid.extract_grid_is_right_handed()
    if ijk_right_handed is None:
        log.warning('ijk handedness not known')
    elif not ijk_right_handed:
        log.warning('ijk axes are left handed; inverted (fake) xyz handedness required')
    assert grid.crs is not None

    if local_coords:
        source_xy_units = grid.crs.xy_units
        source_z_units = grid.crs.z_units
    else:
        if not global_xy_units:
            global_xy_units = grid.crs.xy_units
        if not global_z_units:
            global_z_units = grid.crs.z_units
        if not global_z_increasing_downward:
            log.warning('global z is not increasing with depth as expected by Nexus')
            tm.log_nexus_tm('warning')
        log.info('converting corner points from local to global reference system')
        cp = grid.local_to_global_crs(cp,
                                      crs_uuid = grid.crs.uuid,
                                      global_xy_units = global_xy_units,
                                      global_z_units = global_z_units,
                                      global_z_increasing_downward = global_z_increasing_downward)
        source_xy_units = global_xy_units
        source_z_units = global_z_units

    if not nexus_unit_system:
        guessing_units = grid.crs.z_units if local_coords else global_z_units
        if not guessing_units:
            guessing_units = 'm'
        if guessing_units.startswith('ft'):
            nexus_unit_system = 'ENGLISH'
        elif guessing_units in ['cm', 'mm', 'nm']:
            nexus_unit_system = 'LAB'
        else:
            nexus_unit_system = 'METRIC'
    assert nexus_unit_system in ['METRIC', 'METBAR', 'METKG/CM2', 'ENGLISH', 'LAB']

    if nexus_unit_system.startswith('MET'):
        target_uom = 'm'
    elif nexus_unit_system == 'LAB':
        target_uom = 'cm'
    else:
        target_uom = 'ft'
    wam.convert_lengths(cp[..., :2], source_xy_units, target_uom)
    wam.convert_lengths(cp[..., 2], source_z_units, target_uom)

    log.info(f'writing simulator corner point file: {file_name}; Nexus unit system: {nexus_unit_system}')
    with open(file_name, 'w') as header:
        header.write('! Nexus corner point data written by resqml_grid module\n')
        header.write('! Nexus is a registered trademark of the Halliburton Company\n\n')
        if write_units_keyword:
            header.write(f'{nexus_unit_system}\n\n')
        if write_nx_ny_nz:
            header.write('NX      NY      NZ\n')
            header.write('{0:<7d} {1:<7d} {2:<7d}\n\n'.format(grid.extent_kji[2], grid.extent_kji[1],
                                                              grid.extent_kji[0]))
        if write_rh_keyword_if_needed:
            if ijk_right_handed is None:
                log.warning('unable to determine whether RIGHTHANDED keyword is needed')
            else:
                if local_coords:
                    z_inc_down = grid.z_inc_down()
                    if not z_inc_down:
                        log.warning('local z is not increasing with depth as expected by Nexus')
                        tm.log_nexus_tm('warning')
                else:
                    z_inc_down = global_z_increasing_downward
                xyz_handedness = rqet.xyz_handedness(grid.crs.axis_order, z_inc_down)
                if xyz_handedness == 'unknown':
                    log.warning(
                        'xyz handedness is not known; unable to determine whether RIGHTHANDED keyword is needed')
                else:
                    if ijk_right_handed == (xyz_handedness == 'right'):  # if either both True or both False
                        header.write('RIGHTHANDED\n\n')

    if write_corp_keyword:
        keyword = 'CORP VALUE'
    else:
        keyword = None

    wd.write_array_to_ascii_file(file_name,
                                 corp_extent,
                                 cp.reshape(tuple(corp_extent)),
                                 target_simulator = 'nexus',
                                 keyword = keyword,
                                 columns = 3,
                                 blank_line_after_i_block = False,
                                 blank_line_after_j_block = True,
                                 append = True,
                                 use_binary = use_binary,
                                 binary_only = binary_only,
                                 nan_substitute_value = nan_substitute_value)



"""Class containing resqml stringlookup class."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

# import xml.etree.ElementTree as et

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class StringLookup(BaseResqpy):
    """Class catering for RESQML obj_StringLookupTable objects."""

    resqml_type = "StringTableLookup"

    def __init__(self,
                 parent_model,
                 uuid = None,
                 int_to_str_dict = None,
                 title = None,
                 extra_metadata = None,
                 originator = None):
        """Creates a new string lookup (RESQML obj_StringTableLookup) object.

        arguments:
           parent_model: the model to which this string lookup belongs
           uuid (optional): if present, the uuid for an exising StringTableLookup from which this object is populated
           int_to_str_dict (optional): if present, a dictionary mapping from integers to strings, used to populate the lookup;
              ignored if uuid is present
           title (string, optional): if present, is used as the citation title for the object; ignored if uuid is not None

        returns:
           the new StringLookup object

        :meta common:
        """

        self.min_index = None
        self.max_index = None
        self.str_list = []
        self.str_dict = {}
        self.stored_as_list = False
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)
        if uuid is None:
            self.load_from_dict(int_to_str_dict)

    def _load_from_xml(self):
        root_node = self.root
        for v_node in rqet.list_of_tag(root_node, 'Value'):
            key = rqet.find_tag_int(v_node, 'Key')
            value = rqet.find_tag_text(v_node, 'Value')
            assert key not in self.str_dict, 'key value ' + str(
                key) + ' occurs more than once in string lookup table xml'
            self.str_dict[key] = value
            if self.min_index is None or key < self.min_index:
                self.min_index = key
            if self.max_index is None or key > self.max_index:
                self.max_index = key

    def load_from_dict(self, int_to_str_dict):
        """Sets the contents of this string lookup based on a dict mapping int to str."""

        if int_to_str_dict is None:
            return
        assert len(int_to_str_dict), 'empty dictionary passed to string lookup initialisation'
        self.str_dict = int_to_str_dict.copy()
        self.min_index = min(self.str_dict.keys())
        self.max_index = max(self.str_dict.keys())
        self.set_list_from_dict_conditionally()

    def as_dict(self):
        """Returns the string lookup as a python dictionary."""

        return self.str_dict

    def is_equivalent(self, other):
        """Returns True if this lookup is the same as other (apart from uuid); False otherwise."""

        if other is None:
            return False
        if self is other:
            return True
        if bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.title != other.title or self.min_index != other.min_index or self.max_index != other.max_index:
            return False
        return self.str_dict == other.str_dict

    def set_list_from_dict_conditionally(self):
        """Sets a list copy of the lookup table, which can be indexed directly, if it makes sense to do so."""

        self.str_list = []
        self.stored_as_list = False
        if self.min_index >= 0 and (self.max_index < 50 or 10 * len(self.str_dict) // self.max_index > 8):
            for key in range(self.max_index + 1):
                if key in self.str_dict:
                    self.str_list.append(self.str_dict[key])
                else:
                    self.str_list.append(None)
            self.stored_as_list = True

    def set_string(self, key, value):
        """Sets the string associated with a given integer key."""

        self.str_dict[key] = value
        limits_changed = False
        if self.min_index is None or value < self.min_index:
            self.min_index = value
            limits_changed = True
        if self.max_index is None or value > self.max_index:
            self.max_index = value
            limits_changed = True
        if self.stored_as_list:
            if limits_changed:
                self.set_list_from_dict_conditionally()
            else:
                self.str_list[key] = value

    def get_string(self, key):
        """Returns the string associated with the integer key, or None if not found.

        :meta common:
        """

        if key < self.min_index or key > self.max_index:
            return None
        if self.stored_as_list:
            return self.str_list[key]
        if key not in self.str_dict:
            return None
        return self.str_dict[key]

    def get_list(self):
        """Returns a list of values, sorted by key.

        :meta common:
        """

        if self.stored_as_list:
            return self.str_list
        return list(dict(sorted(list(self.str_dict.items()))).values())

    def length(self):
        """Returns the nominal length of the lookup table.

        :meta common:
        """

        if self.stored_as_list:
            return len(self.str_list)
        return len(self.str_dict)

    def get_index_for_string(self, string):
        """Returns the integer key for the given string (exact match required), or None if not found.

        :meta common:
        """

        if self.stored_as_list:
            try:
                index = self.str_list.index(string)
                return index
            except Exception:
                return None
        if string not in self.str_dict.values():
            return None
        for k, v in self.str_dict.items():
            if v == string:
                return k
        return None

    def create_xml(self, title = None, originator = None, add_as_part = True, reuse = True):
        """Creates an xml node for the string table lookup.

        arguments:
           title (string, optional): if present, overrides the object's title attribute to be used as citation title
           originator (string, optional): if present, used as the citation creator (otherwise login name is used)
           add_as_part (boolean, default True): if True, the property set is added to the model as a part

        :meta common:
        """

        if title:
            self.title = title

        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object

        sl_node = super().create_xml(add_as_part = False, originator = originator)

        for k, v in self.str_dict.items():

            pair_node = rqet.SubElement(sl_node, ns['resqml2'] + 'Value')
            pair_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'StringLookup')
            pair_node.text = rqet.null_xml_text

            key_node = rqet.SubElement(pair_node, ns['resqml2'] + 'Key')
            key_node.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
            key_node.text = str(k)

            value_node = rqet.SubElement(pair_node, ns['resqml2'] + 'Value')
            value_node.set(ns['xsi'] + 'type', ns['xsd'] + 'string')
            value_node.text = str(v)

        if add_as_part:
            self.model.add_part('obj_StringTableLookup', self.uuid, sl_node)

        return sl_node



"""A submodule containing functions relating to grid cell properties."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.vector_utilities as vec
import resqpy.olio.volume as vol
import resqpy.weights_and_measures as wam

import resqpy.grid
import resqpy.grid._defined_geometry as grr_dg


def thickness(grid,
              cell_kji0 = None,
              points_root = None,
              cache_resqml_array = True,
              cache_cp_array = False,
              cache_thickness_array = True,
              property_collection = None):
    """Returns vertical (z) thickness of cell and/or caches thicknesses for all cells.

    arguments:
       cell_kji0 (optional): if present, the (k, j, i) indices of the individual cell for which the
                             thickness is required; zero based indexing
       cache_resqml_array (boolean, default True): If True, the raw points array from the hdf5 file
                             is cached in memory, but only if it is needed to generate the thickness
       cache_cp_array (boolean, default True): If True, an array of corner points is generated and
                             added as an attribute of the grid, with attribute name corner_points, but only
                             if it is needed in order to generate the thickness
       cache_thickness_array (boolean, default False): if True, thicknesses are generated for all cells in
                             the grid and added as an attribute named array_thickness
       property_collection (property:GridPropertyCollection, optional): If not None, this collection
                             is probed for a suitable thickness or cell length property which is used
                             preferentially to calculating thickness; if no suitable property is found,
                             the calculation is made as if the collection were None

    returns:
       float, being the thickness of cell identified by cell_kji0; or numpy float array if cell_kji0 is None

    notes:
       the function can be used to find the thickness of a single cell, or cache thickness for all cells, or both;
       if property_collection is not None, a suitable thickness or cell length property will be used if present;
       if calculated, thickness is defined as z difference between centre points of top and base faces (TVT);
       at present, assumes K increases with same polarity as z; if not, negative thickness will be calculated;
       units of result are implicitly those of z coordinates in grid's coordinate reference system, or units of
       measure of property array if the result is based on a suitable property

    :meta common:
    """

    def __load_from_property(collection):
        if collection is None:
            return None
        parts = collection.selective_parts_list(property_kind = 'thickness', facet_type = 'netgross', facet = 'gross')
        if len(parts) == 1:
            return collection.cached_part_array_ref(parts[0])
        parts = collection.selective_parts_list(property_kind = 'cell length', facet_type = 'direction', facet = 'K')
        if len(parts) == 1:
            return collection.cached_part_array_ref(parts[0])
        parts = collection.selective_parts_list(property_kind = 'thickness')
        if len(parts) == 1 and collection.facet_for_part(parts[0]) is None:
            return collection.cached_part_array_ref(parts[0])
        return None

    # note: this function optionally looks for a suitable thickness property, otherwise calculates from geometry
    # note: for some geometries, thickness might need to be defined as length of vector between -k & +k face centres (TST)
    # todo: give more control over source of data through optional args; offer TST or TVT option
    # todo: if cp array is not already cached, compute directly from points without generating cp
    # todo: cache uom
    assert cache_thickness_array or (cell_kji0 is not None)

    if hasattr(grid, 'array_thickness'):
        if cell_kji0 is None:
            return grid.array_thickness
        return grid.array_thickness[tuple(cell_kji0)]  # could check for nan here and return None

    thick = __load_from_property(property_collection)
    if thick is not None:
        log.debug('thickness array loaded from property')
        if cache_thickness_array:
            grid.array_thickness = thick.copy()
        if cell_kji0 is None:
            return grid.array_thickness
        return grid.array_thickness[tuple(cell_kji0)]  # could check for nan here and return None

    points_root = grid.resolve_geometry_child('Points', child_node = points_root)
    if cache_thickness_array:
        if cache_cp_array:
            grid.corner_points(points_root = points_root, cache_cp_array = True)
        if hasattr(grid, 'array_corner_points'):
            grid.array_thickness = np.abs(
                np.mean(grid.array_corner_points[:, :, :, 1, :, :, 2] - grid.array_corner_points[:, :, :, 0, :, :, 2],
                        axis = (3, 4)))
            if cell_kji0 is None:
                return grid.array_thickness
            return grid.array_thickness[tuple(cell_kji0)]  # could check for nan here and return None
        grid.array_thickness = np.empty(tuple(grid.extent_kji))
        points = grid.point_raw(cache_array = True)  # cache points regardless
        if points is None:
            return None  # geometry not present
        if grid.k_gaps:
            pillar_thickness = points[grid.k_raw_index_array + 1, ..., 2] - points[grid.k_raw_index_array, ..., 2]
        else:
            pillar_thickness = points[1:, ..., 2] - points[:-1, ..., 2]
        if grid.has_split_coordinate_lines:
            pillar_for_col = grid.create_column_pillar_mapping()
            grid.array_thickness = np.abs(
                0.25 *
                (pillar_thickness[:, pillar_for_col[:, :, 0, 0]] + pillar_thickness[:, pillar_for_col[:, :, 0, 1]] +
                 pillar_thickness[:, pillar_for_col[:, :, 1, 0]] + pillar_thickness[:, pillar_for_col[:, :, 1, 1]]))
        else:
            grid.array_thickness = np.abs(0.25 * (pillar_thickness[:, :-1, :-1] + pillar_thickness[:, :-1, 1:] +
                                                  pillar_thickness[:, 1:, :-1] + pillar_thickness[:, 1:, 1:]))
        if cell_kji0 is None:
            return grid.array_thickness
        return grid.array_thickness[tuple(cell_kji0)]  # could check for nan here and return None

    cp = grid.corner_points(cell_kji0 = cell_kji0,
                            points_root = points_root,
                            cache_resqml_array = cache_resqml_array,
                            cache_cp_array = cache_cp_array)
    if cp is None:
        return None
    return abs(np.mean(cp[1, :, :, 2]) - np.mean(cp[0, :, :, 2]))


def volume(grid,
           cell_kji0 = None,
           points_root = None,
           cache_resqml_array = True,
           cache_cp_array = False,
           cache_centre_array = False,
           cache_volume_array = True,
           property_collection = None,
           required_uom = None):
    """Returns bulk rock volume of cell or numpy array of bulk rock volumes for all cells.

    arguments:
       cell_kji0 (optional): if present, the (k, j, i) indices of the individual cell for which the
                             volume is required; zero based indexing
       cache_resqml_array (boolean, default True): If True, the raw points array from the hdf5 file
                             is cached in memory, but only if it is needed to generate the volume
       cache_cp_array (boolean, default False): If True, an array of corner points is generated and
                             added as an attribute of the grid, with attribute name corner_points, but only
                             if it is needed in order to generate the volume
       cache_volume_array (boolean, default False): if True, volumes are generated for all cells in
                             the grid and added as an attribute named array_volume
       property_collection (property:GridPropertyCollection, optional): If not None, this collection
                             is probed for a suitable volume property which is used preferentially
                             to calculating volume; if no suitable property is found,
                             the calculation is made as if the collection were None
       required_uom (str, optional): if present, the RESQML unit of measure (for quantity volume) that
                             the volumes will be returned (and cached) in; if None, the grid's CRS
                             z units cubed will be used

    returns:
       float, being the volume of cell identified by cell_kji0;
       or numpy float array of shape (nk, nj, ni) if cell_kji0 is None

    notes:
       the function can be used to find the volume of a single cell, or cache volumes for all cells, or both;
       if property_collection is not None, a suitable volume property will be used if present;
       if calculated, volume is computed using 6 tetras each with a non-planar bilinear base face;

    :meta common:
    """

    # note: this function optionally looks for a suitable volume property, otherwise calculates from geometry
    # todo: give control over source with optional arguments
    # todo: cache uom

    def __load_vol_from_property(collection):  # returns array and uom
        if collection is None:
            return (None, None)
        parts = collection.selective_parts_list(property_kind = 'rock volume', facet_type = 'netgross', facet = 'gross')
        if len(parts) == 1:
            return (collection.cached_part_array_ref(parts[0]), collection.uom_for_part(parts[0]))
        parts = collection.selective_parts_list(property_kind = 'rock volume')
        if len(parts) == 1 and collection.facet_for_part(parts[0]) is None:
            return (collection.cached_part_array_ref(parts[0]), collection.uom_for_part(parts[0]))
        return (None, None)

    assert (cache_volume_array is not None) or (cell_kji0 is not None)

    required_uom = _get_volume_uom(grid, required_uom)

    if hasattr(grid, 'array_volume'):
        if required_uom != grid.array_volume_uom:
            grid.array_volume = wam.convert_volumes(grid.array_volume, grid.array_volume_uom, required_uom)
            grid.array_volume_uom = required_uom
        if cell_kji0 is None:
            return grid.array_volume
        return grid.array_volume[tuple(cell_kji0)]  # could check for nan here and return None

    vol_array, vol_uom = __load_vol_from_property(property_collection)
    if vol_array is not None:
        if (cache_volume_array or cell_kji0 is None) and (vol_uom != required_uom):
            vol_array = vol_array.copy()
            wam.convert_volumes(vol_array, vol_uom, required_uom)
            vol_uom = required_uom
        if cache_volume_array:
            grid.array_volume = vol_array
            grid.array_volume_uom = vol_uom
        if cell_kji0 is None:
            return vol_array
        if vol_uom == required_uom:
            return vol_array[tuple(cell_kji0)]  # could check for nan here and return None
        return wam.convert(vol_array[tuple(cell_kji0)], vol_uom, required_uom)

    cache_cp_array = cache_cp_array or cell_kji0 is None
    cache_volume_array = cache_volume_array or cell_kji0 is None

    off_hand = grid.off_handed()

    conversion_factor = _get_volume_conversion_factor(grid, required_uom)

    points_root = grid.resolve_geometry_child('Points', child_node = points_root)
    if points_root is None:
        return None  # geometry not present
    centre_array = None
    if cache_volume_array:
        grid.corner_points(points_root = points_root, cache_cp_array = True)
        if cache_centre_array:
            grid.centre_point(cache_centre_array = True)
            centre_array = grid.array_centre_point
        vol_array = vol.tetra_volumes(grid.array_corner_points, centres = centre_array, off_hand = off_hand)
        if conversion_factor is not None:
            vol_array *= conversion_factor
        if cache_volume_array:
            grid.array_volume = vol_array
            grid.array_volume_uom = required_uom
        if cell_kji0 is None:
            return vol_array
        return vol_array[tuple(cell_kji0)]  # could check for nan here and return None

    cp = grid.corner_points(cell_kji0 = cell_kji0,
                            points_root = points_root,
                            cache_resqml_array = cache_resqml_array,
                            cache_cp_array = cache_cp_array)
    if cp is None:
        return None
    v = vol.tetra_cell_volume(cp, off_hand = off_hand)
    return v if conversion_factor is None else conversion_factor * v


def _get_volume_uom(grid, required_uom):
    if required_uom:
        assert required_uom in wam.valid_uoms(quantity = 'volume'), f'invalid volume unit of measure: {required_uom}'
    else:
        linear_uom = grid.crs.z_units
        if linear_uom.startswith('ft'):
            linear_uom = 'ft'
        required_uom = linear_uom + '3'
        if required_uom not in wam.valid_uoms(quantity = 'volume'):
            required_uom = 'm3'
    return required_uom


def _get_volume_conversion_factor(grid, required_uom):
    conversion_factor = None
    if grid.crs.xy_units != grid.crs.z_units or required_uom != grid.crs.z_units + '3':
        conversion_factor = wam.convert(1.0, grid.crs.xy_units, 'm')
        conversion_factor *= conversion_factor
        conversion_factor *= wam.convert(1.0, grid.crs.z_units, 'm')
        if required_uom != 'm3':
            conversion_factor *= wam.convert(1.0, 'm3', required_uom)
    return conversion_factor


def pinched_out(grid,
                cell_kji0 = None,
                tolerance = 0.001,
                points_root = None,
                cache_resqml_array = True,
                cache_cp_array = False,
                cache_thickness_array = False,
                cache_pinchout_array = None):
    """Returns boolean or boolean array indicating whether cell is pinched out.

    Pinched out means cell has a thickness less than tolerance.

    :meta common:
    """

    # note: this function returns a derived object rather than a native resqml object
    # note: returns True for cells without geometry
    # todo: check behaviour in response to NaNs and undefined geometry
    if cache_pinchout_array is None:
        cache_pinchout_array = (cell_kji0 is None)

    if grid.pinchout is not None:
        if cell_kji0 is None:
            return grid.pinchout
        return grid.pinchout[tuple(cell_kji0)]

    if points_root is None:
        points_root = grid.resolve_geometry_child('Points', child_node = points_root)
    #         if points_root is None: return None  # geometry not present

    thick = grid.thickness(
        cell_kji0,
        points_root = points_root,
        cache_resqml_array = cache_resqml_array,
        cache_cp_array = cache_cp_array,  # deprecated
        cache_thickness_array = cache_thickness_array or cache_pinchout_array)
    if cache_pinchout_array:
        grid.pinchout = np.where(np.isnan(grid.array_thickness), True, np.logical_not(grid.array_thickness > tolerance))
        if cell_kji0 is None:
            return grid.pinchout
        return grid.pinchout[tuple(cell_kji0)]
    if thick is not None:
        return thick <= tolerance
    return None


def cell_inactive(grid, cell_kji0, pv_array = None, pv_tol = 0.01):
    """Returns True if the cell is inactive."""

    if grid.inactive is not None:
        return grid.inactive[tuple(cell_kji0)]
    grid.extract_inactive_mask()
    if grid.inactive is not None:
        return grid.inactive[tuple(cell_kji0)]
    if pv_array is not None:  # fabricate an inactive mask from pore volume data
        grid.inactive = not (pv_array > pv_tol)  # NaN in pv array will end up inactive
        return grid.inactive[tuple(cell_kji0)]
    return ((not grr_dg.cell_geometry_is_defined(grid, cell_kji0 = cell_kji0)) or
            grid.pinched_out(cell_kji0, cache_pinchout_array = True))


def interface_vector(grid, cell_kji0, axis, points_root = None, cache_resqml_array = True, cache_cp_array = False):
    """Returns an xyz vector between centres of an opposite pair of faces of the cell (or vectors for all cells).

    note:
        units are implicitly those of the grid's crs; differing xy & z units would imply that the direction of the
        vector is not a true direction
    """

    face_0_centre = grid.face_centre(cell_kji0,
                                     axis,
                                     0,
                                     points_root = points_root,
                                     cache_resqml_array = cache_resqml_array,
                                     cache_cp_array = cache_cp_array)
    face_1_centre = grid.face_centre(cell_kji0,
                                     axis,
                                     1,
                                     points_root = points_root,
                                     cache_resqml_array = cache_resqml_array,
                                     cache_cp_array = cache_cp_array)
    return face_1_centre - face_0_centre


def interface_length(grid,
                     cell_kji0,
                     axis,
                     points_root = None,
                     cache_resqml_array = True,
                     cache_cp_array = False,
                     required_uom = None):
    """Returns the length between centres of an opposite pair of faces of the cell.

    note:
       if required_uom is not specified, units of returned length are the grid's crs xy units
    """

    assert cell_kji0 is not None
    v = grid.interface_vector(cell_kji0,
                              axis,
                              points_root = points_root,
                              cache_resqml_array = cache_resqml_array,
                              cache_cp_array = cache_cp_array)
    if grid.crs.xy_units != grid.crs.z_units:
        v[2] = wam.convert_lengths(v[2], grid.crs.z_units, grid.crs.xy_units)
    if required_uom:
        wam.convert_lengths(v, grid.crs.xy_units, required_uom)
    return vec.naive_length(v)


def interface_vectors_kji(grid, cell_kji0, points_root = None, cache_resqml_array = True, cache_cp_array = False):
    """Returns 3 interface centre point difference vectors for axes k, j, i.

    note:
        units are implicitly those of the grid's crs; differing xy & z units would imply that the direction of the
        vectors are not true directions
    """

    result = np.zeros((3, 3))
    for axis in range(3):
        result[axis] = grid.interface_vector(cell_kji0,
                                             axis,
                                             points_root = points_root,
                                             cache_resqml_array = cache_resqml_array,
                                             cache_cp_array = cache_cp_array)
    return result


def interface_lengths_kji(grid,
                          cell_kji0,
                          points_root = None,
                          cache_resqml_array = True,
                          cache_cp_array = False,
                          required_uom = None):
    """Returns 3 interface centre point separation lengths for axes k, j, i.

    note:
       if required_uom is not specified, units of returned lengths are the grid's crs xy units
    """
    result = np.zeros(3)
    for axis in range(3):
        result[axis] = grid.interface_length(cell_kji0,
                                             axis,
                                             points_root = points_root,
                                             cache_resqml_array = cache_resqml_array,
                                             cache_cp_array = cache_cp_array,
                                             required_uom = required_uom)
    return result


def poly_line_for_cell(grid, cell_kji0, vertical_ref = 'top'):
    """Returns a numpy array of shape (4, 3) being the 4 corners.

    Corners are in order J-I-, J-I+, J+I+, J+I-; from the top or base face.
    """
    if vertical_ref == 'top':
        kp = 0
    elif vertical_ref == 'base':
        kp = 1
    else:
        raise ValueError('vertical reference not catered for: ' + vertical_ref)
    poly = np.empty((4, 3))
    cp = grid.corner_points(cell_kji0 = cell_kji0)
    if cp is None:
        return None
    poly[0] = cp[kp, 0, 0]
    poly[1] = cp[kp, 0, 1]
    poly[2] = cp[kp, 1, 1]
    poly[3] = cp[kp, 1, 0]
    return poly



"""Class for a collection of well interval properties"""

# NB. This class does not appear to be well formed and is a candidate for deprecation

import logging

log = logging.getLogger(__name__)

import pandas as pd

import resqpy.property
import resqpy.property.property_collection as rqp_pc
import resqpy.property.well_interval_property as rqp_wip
import resqpy.property.property_common as rqp_c


class WellIntervalPropertyCollection(rqp_pc.PropertyCollection):
    """Class for RESQML property collection for a WellboreFrame for interval or blocked well logs"""

    def __init__(self, frame = None, property_set_root = None, realization = None):
        """Creates a new property collection related to interval or blocked well logs and a wellbore frame."""

        super().__init__(support = frame, property_set_root = property_set_root, realization = realization)

    def logs(self):
        """Generator that yields component Interval log or Blocked well log objects."""

        return (rqp_wip.WellIntervalProperty(collection = self, part = part) for part in self.parts())

    def to_pandas(self, include_units = False):
        """Returns a dataframe with a column for each well log included in the collection."""

        cell_indices = [rqp_c.return_cell_indices(i, self.support.cell_indices) for i in self.support.cell_grid_link]
        data = {}
        for log in self.logs():
            col_name = log.name
            values = log.values()
            data[col_name] = values
        df = pd.DataFrame(data = data, index = cell_indices)
        return df



"""Submodule containing the IntervalsInfo class."""


class IntervalsInfo:
    """A class representing intervals"""

    def __init__(self):
        """Declaring the properties of the Intervals Info class"""
        self.child_count_per_interval = None
        self.parent_count_per_interval = None
        self.child_cell_weights = None



"""High level coarsened grid function."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.crs as rqc
import resqpy.derived_model
import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.fine_coarse as fc
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp

import resqpy.derived_model._common as rqdm_c


def coarsened_grid(epc_file,
                   source_grid,
                   fine_coarse,
                   inherit_properties = False,
                   inherit_realization = None,
                   inherit_all_realizations = False,
                   set_parent_window = None,
                   infill_missing_geometry = True,
                   new_grid_title = None,
                   new_epc_file = None):
    """Generates a coarsened version of an unsplit source grid, optionally inheriting properties.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): if None, the epc_file is loaded and it should contain one ijk grid object
          (or one 'ROOT' grid) which is used as the source grid
       fine_coarse (resqpy.olio.fine_coarse.FineCoarse object): the mapping between cells in the fine (source) and
          coarse (output) grids
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid, with values upscaled or sampled
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       set_parent_window (boolean or str, optional): if True or 'parent', the coarsened grid has its parent window attribute
          set; if False, the parent window is not set; if None, the default will be True if new_epc_file is None or False
          otherwise; if 'grandparent' then an intervening parent window with no refinement or coarsening will be skipped
          and its box used in the parent window for the new grid, relating directly to the original grid
       infill_missing_geometry (boolean, default True): if True, an attempt is made to generate grid geometry in the
          source grid wherever it is undefined; if False, any undefined geometry will result in an assertion failure
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the refined grid (& crs)

    returns:
       new grid object being the coarsened grid; the epc and hdf5 files are written to as an intentional side effect

    note:
       this function coarsens an entire grid; to coarsen a local area of a grid, first use the extract_box function
       and then use this function on the extracted grid; in such a case, using a value of 'grandparent' for the
       set_parent_window argument will relate the coarsened grid back to the original
    """

    new_epc_file, model, source_grid = _establish_files_and_model(epc_file, new_epc_file, source_grid)

    if set_parent_window is None:
        set_parent_window = (new_epc_file is None)
    assert fine_coarse is not None and isinstance(fine_coarse, fc.FineCoarse)

    assert not source_grid.has_split_coordinate_lines, 'coarsening only available for unsplit grids: use other functions to heal faults first'

    if infill_missing_geometry and (not source_grid.geometry_defined_for_all_cells() or
                                    not source_grid.geometry_defined_for_all_pillars()):
        log.debug('attempting infill of geometry missing in source grid')
        source_grid.set_geometry_is_defined(treat_as_nan = None,
                                            treat_dots_as_nan = True,
                                            complete_partial_pillars = True,
                                            nullify_partial_pillars = False,
                                            complete_all = True)

    assert source_grid.geometry_defined_for_all_pillars(), 'coarsening requires geometry to be defined for all pillars'
    assert source_grid.geometry_defined_for_all_cells(), 'coarsening requires geometry to be defined for all cells'
    assert not source_grid.k_gaps, 'coarsening of grids with k gaps not currently supported'

    assert tuple(fine_coarse.fine_extent_kji) == tuple(source_grid.extent_kji),  \
           'fine_coarse mapping fine extent does not match that of source grid'
    fine_coarse.assert_valid()

    source_grid.cache_all_geometry_arrays()
    source_points = source_grid.points_ref().reshape((source_grid.nk + 1), (source_grid.nj + 1) * (source_grid.ni + 1),
                                                     3)

    # create a new, empty grid object
    grid = grr.Grid(model)

    # inherit attributes from source grid
    grid.grid_representation = 'IjkGrid'
    grid.extent_kji = fine_coarse.coarse_extent_kji
    grid.nk, grid.nj, grid.ni = grid.extent_kji[0], grid.extent_kji[1], grid.extent_kji[2]
    grid.k_direction_is_down = source_grid.k_direction_is_down
    grid.grid_is_right_handed = source_grid.grid_is_right_handed
    grid.pillar_shape = source_grid.pillar_shape
    grid.has_split_coordinate_lines = False
    grid.split_pillars_count = None
    # inherit the coordinate reference system used by the grid geometry
    grid.crs_uuid = source_grid.crs_uuid
    if source_grid.model is not model:
        model.duplicate_node(source_grid.model.root_for_uuid(grid.crs_uuid), add_as_part = True)
    grid.crs = rqc.Crs(model, grid.crs_uuid)

    coarsened_points = np.empty(
        (grid.nk + 1, (grid.nj + 1) * (grid.ni + 1), 3))  # note: gets reshaped after being populated

    k_ratio_constant = fine_coarse.constant_ratios[0]
    if k_ratio_constant:
        k_indices = None
    else:
        k_indices = np.empty(grid.nk + 1, dtype = int)
        k_indices[0] = 0
        for k in range(grid.nk):
            k_indices[k + 1] = k_indices[k] + fine_coarse.vector_ratios[0][k]
        assert k_indices[-1] == source_grid.nk

    for cjp in range(grid.nj + 1):
        for cji in range(grid.ni + 1):
            natural_coarse_pillar = cjp * (grid.ni + 1) + cji
            natural_fine_pillar = fine_coarse.fine_for_coarse_natural_pillar_index(natural_coarse_pillar)
            if k_ratio_constant:
                coarsened_points[:, natural_coarse_pillar, :] = source_points[0:source_grid.nk + 1:k_ratio_constant,
                                                                              natural_fine_pillar, :]
            else:
                coarsened_points[:, natural_coarse_pillar, :] = source_points[k_indices, natural_fine_pillar, :]

    grid.points_cached = coarsened_points.reshape(((grid.nk + 1), (grid.nj + 1), (grid.ni + 1), 3))

    grid.geometry_defined_for_all_pillars_cached = True
    grid.geometry_defined_for_all_cells_cached = True
    grid.array_cell_geometry_is_defined = np.full(tuple(grid.extent_kji), True, dtype = bool)

    collection = None
    if inherit_properties:
        source_collection = source_grid.extract_property_collection()
        if source_collection is not None:
            collection = rqp.GridPropertyCollection()
            collection.set_grid(grid)
            collection.extend_imported_list_copying_properties_from_other_grid_collection(
                source_collection,
                coarsening = fine_coarse,
                realization = inherit_realization,
                copy_all_realizations = inherit_all_realizations)

    _set_parent_window_in_grid(set_parent_window, source_grid, grid, fine_coarse)

    # write grid
    if new_grid_title is None or len(new_grid_title) == 0:
        new_grid_title = 'grid coarsened from ' + str(rqet.citation_title_for_node(source_grid.root))

    model.h5_release()
    if new_epc_file:
        rqdm_c._write_grid(new_epc_file,
                           grid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'w')
    else:
        ext_uuid, _ = model.h5_uuid_and_path_for_node(rqet.find_nested_tags(source_grid.root, ['Geometry', 'Points']),
                                                      'Coordinates')
        rqdm_c._write_grid(epc_file,
                           grid,
                           ext_uuid = ext_uuid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'a')

    return grid


def _set_parent_window_in_grid(set_parent_window, source_grid, grid, fine_coarse):
    if set_parent_window:
        pw_grid_uuid = source_grid.uuid
        if isinstance(set_parent_window, str):
            if set_parent_window == 'grandparent':
                assert fine_coarse.within_fine_box is None or (np.all(fine_coarse.within_fine_box[0] == 0) and
                                                               np.all(fine_coarse.within_fine_box[1]) == source_grid.extent_kji - 1),  \
                   'attempt to set grandparent window for grid when parent window is present'
                source_fine_coarse = source_grid.parent_window
                if source_fine_coarse is not None and (source_fine_coarse.within_fine_box is not None or
                                                       source_fine_coarse.within_coarse_box is not None):
                    assert source_fine_coarse.fine_extent_kji == source_fine_coarse.coarse_extent_kji, 'parentage involves refinement or coarsening'
                    if source_fine_coarse.within_fine_box is not None:
                        fine_coarse.within_fine_box = source_fine_coarse.within_fine_box
                    else:
                        fine_coarse.within_fine_box = source_fine_coarse.within_coarse_box
                    pw_grid_uuid = bu.uuid_from_string(
                        rqet.find_nested_tags_text(source_grid.root, ['ParentWindow', 'ParentGrid', 'UUID']))
            else:
                assert set_parent_window == 'parent', 'set_parent_window value not recognized: ' + set_parent_window
        grid.set_parent(pw_grid_uuid, False, fine_coarse)


def _establish_files_and_model(epc_file, new_epc_file, source_grid):
    assert epc_file or new_epc_file, 'epc file name not specified'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    assert epc_file or source_grid is not None, 'neither epc file name nor source grid supplied'
    if epc_file:
        model = rq.Model(epc_file)
        if source_grid is None:
            source_grid = model.grid()  # requires there to be exactly one grid in model (or one named ROOT)
    else:
        model = source_grid.model
    assert source_grid.grid_representation == 'IjkGrid'
    assert model is not None
    return new_epc_file, model, source_grid



"""Model class, roughly equivalent to a RESQML epc file."""

__all__ = ['Model', 'ModelContext', 'new_model']

from ._model import Model, new_model
from ._context import ModelContext

# Set "module" attribute of all public objects to this path. Fixes issue #310
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""Submodule containing functions for attribute extraction for a property collection."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import numpy as np
import numpy.ma as ma

import resqpy
import resqpy.property as rqp
import resqpy.olio.xml_et as rqet
import resqpy.olio.uuid as bu
import resqpy.property.property_kind as rqpk


def _min_max_of_cached_array(collection, cached_name, cached_array, null_value, discrete):
    collection.__dict__[cached_name] = cached_array
    if discrete:
        if null_value is None:
            min_value = np.min(cached_array)
            max_value = np.max(cached_array)
        else:
            zorro = collection.masked_array(cached_array, exclude_value = null_value)
            min_value = np.min(zorro)
            max_value = np.max(zorro)
            if min_value is ma.masked:
                min_value = None
            if max_value is ma.masked:
                max_value = None
    else:
        min_value = np.nanmin(cached_array)
        max_value = np.nanmax(cached_array)
        if np.isnan(min_value):
            min_value = None
        if np.isnan(max_value):
            max_value = None
    return min_value, max_value


def _get_property_kind_uuid(collection, property_kind_uuid, property_kind, uom, discrete):
    if property_kind_uuid is None:
        pk_parts_list = collection.model.parts_list_of_type('PropertyKind')
        for part in pk_parts_list:
            if collection.model.citation_title_for_part(part) == property_kind:
                property_kind_uuid = collection.model.uuid_for_part(part)
                break
        if property_kind_uuid is None:
            # create local property kind object and fetch uuid
            lpk = rqpk.PropertyKind(collection.model,
                                    title = property_kind,
                                    example_uom = uom,
                                    parent_property_kind = 'discrete' if discrete else 'continuous')
            lpk.create_xml()
            property_kind_uuid = lpk.uuid
    return property_kind_uuid


def _get_property_type_details(collection, discrete, string_lookup_uuid, points):
    if discrete:
        if string_lookup_uuid is None:
            collection.d_or_c_text = 'Discrete'

        else:
            collection.d_or_c_text = 'Categorical'
        collection.xsd_type = 'integer'
        collection.hdf5_type = 'IntegerHdf5Array'
    elif points:
        collection.d_or_c_text = 'Points'
        collection.xsd_type = 'double'
        collection.hdf5_type = 'Point3dHdf5Array'
        collection.null_value = None
    else:
        collection.d_or_c_text = 'Continuous'
        collection.xsd_type = 'double'
        collection.hdf5_type = 'DoubleHdf5Array'
        collection.null_value = None


def _get_property_array_min_max_value(collection, property_array, const_value, discrete, min_value, max_value,
                                      categorical, null_value):
    if const_value is not None:
        return _get_property_array_min_max_const(const_value, collection.null_value, min_value, max_value, discrete)
    elif property_array is not None:
        return _get_property_array_min_max_array(property_array, min_value, max_value, discrete, categorical,
                                                 null_value)
    return None, None


def _get_property_array_min_max_const(const_value, null_value, min_value, max_value, discrete):
    if (discrete and const_value != null_value) or (not discrete and not np.isnan(const_value)):
        if min_value is None:
            min_value = const_value
        if max_value is None:
            max_value = const_value
    return min_value, max_value


def _get_property_array_min_max_array(property_array, min_value, max_value, discrete, categorical, null_value):
    if discrete:
        if categorical:
            min_value, max_value = _get_property_array_min_max_array_categorical(property_array, min_value, max_value,
                                                                                 null_value)
        else:
            min_value, max_value = _get_property_array_min_max_array_discrete(property_array, min_value, max_value)
    else:
        min_value, max_value = _get_property_array_min_max_array_continuous(property_array, min_value, max_value)

    return min_value, max_value


def _get_property_array_min_max_array_discrete(property_array, min_value, max_value):
    if min_value is None:
        try:
            min_value = int(property_array.min())
        except Exception:
            min_value = None
            log.warning('no xml minimum value set for discrete property')
    if max_value is None:
        try:
            max_value = int(property_array.max())
        except Exception:
            max_value = None
            log.warning('no xml maximum value set for discrete property')
    return min_value, max_value


def _get_property_array_min_max_array_categorical(property_array, min_value, max_value, null_value):
    if min_value is None or max_value is None:
        unique_values = np.unique(property_array)
        if null_value is not None and len(unique_values):
            if unique_values[0] == null_value:
                unique_values = unique_values[1:]
            elif unique_values[-1] == null_value:
                unique_values = unique_values[:-1]
    if min_value is None:
        if len(unique_values):
            min_value = unique_values[0]
        else:
            log.warning('no xml minimum value set for categorical property')
    if max_value is None:
        if len(unique_values):
            max_value = unique_values[-1]
        else:
            log.warning('no xml maximum value set for categorical property')
    return min_value, max_value


def _get_property_array_min_max_array_continuous(property_array, min_value, max_value):
    if min_value is None or max_value is None:
        all_nan = np.all(np.isnan(property_array))
    if min_value is None and not all_nan:
        min_value = np.nanmin(property_array)
        if np.isnan(min_value) or min_value is ma.masked:
            min_value = None
    if max_value is None and not all_nan:
        max_value = np.nanmax(property_array)
        if np.isnan(max_value) or max_value is ma.masked:
            max_value = None
    return min_value, max_value


def _find_single_part(collection, kind, realization):
    try:
        part = collection.singleton(realization = realization, property_kind = kind)
    except Exception:
        log.error(f'problem with {kind} (more than one array present?)')
        part = None
    return part


def _get_single_perm_ijk_parts(collection, perms, share_perm_parts, perm_k_mode, perm_k_ratio, ntg_part):
    if perms.number_of_parts() == 1:
        return _get_single_perm_ijk_parts_one(perms, share_perm_parts)
    else:
        perm_i_part = _get_single_perm_ijk_for_direction(perms, 'I')
        perm_j_part = _get_single_perm_ijk_for_direction(perms, 'J')
        if perm_j_part is None and share_perm_parts:
            perm_j_part = perm_i_part
        elif perm_i_part is None and share_perm_parts:
            perm_i_part = perm_j_part
        perm_k_part = _get_single_perm_ijk_for_direction(perms, 'K')
        if perm_k_part is None:
            assert perm_k_mode in [None, 'none', 'shared', 'ratio', 'ntg', 'ntg squared']
            # note: could switch ratio mode to shared if perm_k_ratio is 1.0
            if perm_k_mode is None or perm_k_mode == 'none':
                pass
            else:
                perm_k_part = _get_perm_k_part(collection, perms, perm_k_mode, share_perm_parts, perm_i_part,
                                               perm_j_part, perm_k_ratio, ntg_part)
    return perm_i_part, perm_j_part, perm_k_part


def _get_perm_k_part(collection, perms, perm_k_mode, share_perm_parts, perm_i_part, perm_j_part, perm_k_ratio,
                     ntg_part):
    if perm_k_mode == 'shared':
        if share_perm_parts:
            perm_k_part = perm_i_part
    elif perm_i_part is not None:
        log.info('generating K permeability data using mode ' + str(perm_k_mode))
        if perm_j_part is not None and perm_j_part != perm_i_part:
            # generate root mean square of I & J permeabilities to use as horizontal perm
            kh = np.sqrt(perms.cached_part_array_ref(perm_i_part) * perms.cached_part_array_ref(perm_j_part))
        else:  # use I permeability as horizontal perm
            kh = perms.cached_part_array_ref(perm_i_part)
        kv = kh * perm_k_ratio
        if ntg_part is not None:
            if perm_k_mode == 'ntg':
                kv *= collection.cached_part_array_ref(ntg_part)
            elif perm_k_mode == 'ntg squared':
                ntg = collection.cached_part_array_ref(ntg_part)
                kv *= ntg * ntg
        kv_collection = rqp.PropertyCollection()
        kv_collection.set_support(support_uuid = collection.support_uuid, model = collection.model)
        kv_collection.add_cached_array_to_imported_list(kv,
                                                        'derived from horizontal perm with mode ' + str(perm_k_mode),
                                                        'KK',
                                                        discrete = False,
                                                        uom = 'mD',
                                                        time_index = None,
                                                        null_value = None,
                                                        property_kind = 'permeability rock',
                                                        facet_type = 'direction',
                                                        facet = 'K',
                                                        realization = perms.realization_for_part(perm_i_part),
                                                        indexable_element = perms.indexable_for_part(perm_i_part),
                                                        count = 1,
                                                        points = False)
        collection.model.h5_release()
        kv_collection.write_hdf5_for_imported_list()
        kv_collection.create_xml_for_imported_list_and_add_parts_to_model()
        collection.inherit_parts_from_other_collection(kv_collection)
        perm_k_part = kv_collection.singleton()
    return perm_k_part


def _get_single_perm_ijk_parts_one(perms, share_perm_parts):
    perm_i_part = perms.singleton()
    perm_j_part = perm_k_part = None
    if share_perm_parts:
        perm_j_part = perm_k_part = perm_i_part
    elif perms.facet_type_for_part(perm_i_part) == 'direction':
        direction = perms.facet_for_part(perm_i_part)
        if direction == 'J':
            perm_j_part = perm_i_part
            perm_i_part = None
        elif direction == 'IJ':
            perm_j_part = perm_i_part
        elif direction == 'K':
            perm_k_part = perm_i_part
            perm_i_part = None
        elif direction == 'IJK':
            perm_j_part = perm_k_part = perm_i_part
    return perm_i_part, perm_j_part, perm_k_part


def _get_single_perm_ijk_for_direction(perms, direction):
    facet_options, title_options = _get_facet_title_options_for_direction(direction)

    try:
        part = None
        for facet_op in facet_options:
            if not part:
                part = perms.singleton(facet_type = 'direction', facet = facet_op)
        if not part:
            for title in title_options:
                if not part:
                    part = perms.singleton(citation_title = title)
        if not part:
            log.error(f'unable to discern which rock permeability to use for {direction} direction')
    except Exception:
        log.error(f'problem with permeability data (more than one {direction} direction array present?)')
        part = None
    return part


def _get_facet_title_options_for_direction(direction):
    if direction == 'I':
        facet_options = ['I', 'IJ', 'IJK']
        title_options = ['KI', 'PERMI', 'KX', 'PERMX']
    elif direction == 'J':
        facet_options = ['J', 'IJ', 'IJK']
        title_options = ['KJ', 'PERMJ', 'KY', 'PERMY']
    else:
        facet_options = ['K', 'IJK']
        title_options = ['KK', 'PERMK', 'KZ', 'PERMZ']
    return facet_options, title_options


def _part_direction(collection, part):
    facet_t = collection.facet_type_for_part(part)
    if facet_t is None or facet_t != 'direction':
        return None
    return collection.facet_for_part(part)


def _cached_part_array_ref_get_node_points(part_node, dtype):
    patch_list = rqet.list_of_tag(part_node, 'PatchOfPoints')
    assert len(patch_list) == 1  # todo: handle more than one patch of points
    first_values_node = rqet.find_tag(patch_list[0], 'Points')
    if first_values_node is None:
        return None  # could treat as fatal error
    if dtype is None:
        dtype = 'float'
    else:
        assert dtype in ['float', float, np.float32, np.float64]
    tag = 'Coordinates'
    return first_values_node, tag, dtype


def _cached_part_array_ref_get_node_values(part_node, dtype):
    patch_list = rqet.list_of_tag(part_node, 'PatchOfValues')
    assert len(patch_list) == 1  # todo: handle more than one patch of values
    first_values_node = rqet.find_tag(patch_list[0], 'Values')
    if first_values_node is None:
        return None  # could treat as fatal error
    if dtype is None:
        array_type = rqet.node_type(first_values_node)
        assert array_type is not None
        if array_type == 'DoubleHdf5Array':
            dtype = 'float'
        elif array_type == 'IntegerHdf5Array':
            dtype = 'int'
        elif array_type == 'BooleanHdf5Array':
            dtype = 'bool'
        else:
            raise ValueError('array type not catered for: ' + str(array_type))
    tag = 'Values'
    return first_values_node, tag, dtype


def _process_imported_property_get_add_min_max(points, property_kind, string_lookup_uuid, local_property_kind_uuid,
                                               is_bool):
    if points or property_kind == 'categorical':
        add_min_max = True
    elif property_kind == 'discrete' and is_bool:
        add_min_max = False
    elif local_property_kind_uuid is not None and string_lookup_uuid is not None:
        add_min_max = False
    else:
        add_min_max = True
    return add_min_max


def _normalized_part_array_apply_discrete_cycle(discrete_cycle, p_array, min_value, max_value):
    if 'int' in str(
            p_array.dtype) and discrete_cycle is not None:  # could use continuous flag in metadata instead of dtype
        p_array = p_array % discrete_cycle
        min_value = 0
        max_value = discrete_cycle - 1
    elif str(p_array.dtype).startswith('bool'):
        min_value = int(min_value)
        max_value = int(max_value)
    return min_value, max_value, p_array


def _normalized_part_array_nan_if_masked(min_value, max_value, masked):
    min_value = float(min_value)  # will return np.ma.masked if all values are masked out
    if masked and min_value is ma.masked:
        min_value = np.nan
    max_value = float(max_value)
    if masked and max_value is ma.masked:
        max_value = np.nan
    return min_value, max_value


def _normalized_part_array_use_logarithm(min_value, n_prop, masked):
    if min_value <= 0.0:
        n_prop[:] = np.where(n_prop < 0.0001, 0.0001, n_prop)
    n_prop = np.log10(n_prop)
    min_value = np.nanmin(n_prop)
    max_value = np.nanmax(n_prop)
    if masked:
        if min_value is ma.masked:
            min_value = np.nan
        if max_value is ma.masked:
            max_value = np.nan
    return n_prop, min_value, max_value


def _normalized_part_array_fix_zero_at(min_value, max_value, n_prop, fix_zero_at):
    if fix_zero_at <= 0.0:
        if min_value < 0.0:
            n_prop[:] = np.where(n_prop < 0.0, 0.0, n_prop)
        min_value = 0.0
    elif fix_zero_at >= 1.0:
        if max_value > 0.0:
            n_prop[:] = np.where(n_prop > 0.0, 0.0, n_prop)
        max_value = 0.0
    else:
        upper_scaling = max_value / (1.0 - fix_zero_at)
        lower_scaling = -min_value / fix_zero_at
        if upper_scaling >= lower_scaling:
            min_value = -upper_scaling * fix_zero_at
            n_prop[:] = np.where(n_prop < min_value, min_value, n_prop)
        else:
            max_value = lower_scaling * (1.0 - fix_zero_at)
            n_prop[:] = np.where(n_prop > max_value, max_value, n_prop)
    return min_value, max_value, n_prop


def _supporting_shape_grid(support, indexable_element, direction):
    if indexable_element is None or indexable_element == 'cells':
        shape_list = [support.nk, support.nj, support.ni]
    elif indexable_element == 'columns':
        shape_list = [support.nj, support.ni]
    elif indexable_element == 'layers':
        shape_list = [support.nk]
    elif indexable_element == 'faces':
        shape_list = _supporting_shape_grid_faces(direction, support)
    elif indexable_element == 'column edges':
        # I edges first; include outer edges
        shape_list = [(support.nj * (support.ni + 1)) + ((support.nj + 1) * support.ni)]
    elif indexable_element == 'edges per column':
        shape_list = [support.nj, support.ni, 4]  # assume I-, J+, I+, J- ordering
    elif indexable_element == 'faces per cell':
        shape_list = [support.nk, support.nj, support.ni, 6]  # assume K-, K+, J-, I+, J+, I- ordering
        # TODO: resolve ordering of edges and make consistent with maps code (edges per column) and fault module (gcs faces)
    elif indexable_element == 'nodes per cell':
        # kp, jp, ip within each cell; todo: check RESQML shaping
        shape_list = [support.nk, support.nj, support.ni, 2, 2, 2]
    elif indexable_element == 'nodes':
        shape_list = _supporting_shape_grid_nodes(support)
    elif indexable_element == 'J0':
        shape_list = [support.nk, support.ni]
    elif indexable_element == 'I0':
        shape_list = [support.nk, support.nj]
    return shape_list


def _supporting_shape_grid_faces(direction, support):
    assert direction is not None and direction.upper() in 'IJK'
    axis = 'KJI'.index(direction.upper())
    shape_list = [support.nk, support.nj, support.ni]
    shape_list[axis] += 1  # note: properties for grid faces include outer faces
    return shape_list


def _supporting_shape_grid_nodes(support):
    assert not support.k_gaps, 'indexable element of nodes not currently supported for grids with K gaps'
    if support.has_split_coordinate_lines:
        pillar_count = (support.nj + 1) * (support.ni + 1) + support.split_pillars_count
        shape_list = [support.nk + 1, pillar_count]
    else:
        shape_list = [support.nk + 1, support.nj + 1, support.ni + 1]
    return shape_list


def _supporting_shape_wellboreframe(support, indexable_element):
    if indexable_element is None or indexable_element == 'nodes':
        shape_list = [support.node_count]
    elif indexable_element == 'intervals':
        shape_list = [support.node_count - 1]
    return shape_list


def _supporting_shape_wellboremarkerframe(support, indexable_element):
    if indexable_element is None or indexable_element == 'nodes':
        shape_list = [support.node_count]
    elif indexable_element == 'intervals':
        shape_list = [support.node_count - 1]
    return shape_list


def _supporting_shape_blockedwell(support, indexable_element):
    if indexable_element is None or indexable_element == 'intervals':
        shape_list = [support.node_count - 1]  # all intervals, including unblocked
    elif indexable_element == 'nodes':
        shape_list = [support.node_count]
    elif indexable_element == 'cells':
        shape_list = [support.cell_count]  # ie. blocked intervals only
    return shape_list


def _supporting_shape_mesh(support, indexable_element):
    if indexable_element is None or indexable_element == 'cells' or indexable_element == 'columns':
        shape_list = [support.nj - 1, support.ni - 1]
    elif indexable_element == 'nodes':
        shape_list = [support.nj, support.ni]
    return shape_list


def _supporting_shape_surface(support, indexable_element):
    if indexable_element is None or indexable_element == 'faces':
        shape_list = [support.triangle_count()]
    elif indexable_element == 'nodes':
        shape_list = [support.node_count()]
    return shape_list


def _supporting_shape_gridconnectionset(support, indexable_element):
    if indexable_element is None or indexable_element == 'faces':
        shape_list = [support.count]
    return shape_list


def _supporting_shape_other(support, indexable_element):
    if indexable_element is None or indexable_element == 'cells':
        shape_list = [support.cell_count]
    elif indexable_element == 'faces per cell':
        support.cache_all_geometry_arrays()
        shape_list = [len(support.faces_per_cell)]
    return shape_list, support


def _realizations_array_ref_get_r_extent(fill_missing, r_list):
    if fill_missing:
        r_extent = r_list[-1] + 1
    else:
        r_extent = len(r_list)
    return r_extent


def _get_indexable_element(indexable_element, support_type):
    if indexable_element is None:
        if support_type in [
                'obj_IjkGridRepresentation', 'obj_BlockedWellboreRepresentation', 'obj_Grid2dRepresentation',
                'obj_UnstructuredGridRepresentation'
        ]:
            indexable_element = 'cells'
        elif support_type in ['obj_WellboreFrameRepresentation', 'obj_WellboreMarkerFrameRepresentation']:
            indexable_element = 'nodes'  # note: could be 'intervals'
        elif support_type in ['obj_GridConnectionSetRepresentation', 'obj_TriangulatedSetRepresentation']:
            indexable_element = 'faces'
        else:
            raise Exception('indexable element unknown for unsupported supporting representation object')
    return indexable_element


def _cached_part_array_ref_get_array(collection, part, dtype, model, cached_array_name, use_pack):
    const_value = collection.constant_value_for_part(part)
    if const_value is None:
        _cached_part_array_ref_const_none(collection, part, dtype, model, cached_array_name, use_pack)
    else:
        _cached_part_array_ref_const_notnone(collection, part, const_value, cached_array_name)


def _cached_part_array_ref_const_none(collection, part, dtype, model, cached_array_name, use_pack):
    part_node = collection.node_for_part(part)
    if part_node is None:
        return None
    if collection.points_for_part(part):
        first_values_node, tag, dtype = _cached_part_array_ref_get_node_points(part_node, dtype)
    else:
        first_values_node, tag, dtype = _cached_part_array_ref_get_node_values(part_node, dtype)

    # the required shape is required if a bool array may need to be unpacked from bits
    required_shape = None
    str_dtype = str(dtype)
    if use_pack and ('bool' in str_dtype or 'int8' in str_dtype):
        required_shape = collection.supporting_shape(indexable_element = collection.indexable_for_part(part),
                                                     direction = _part_direction(collection, part))

    h5_key_pair = model.h5_uuid_and_path_for_node(first_values_node, tag = tag)
    if h5_key_pair is None:
        return None
    model.h5_array_element(h5_key_pair,
                           index = None,
                           cache_array = True,
                           object = collection,
                           array_attribute = cached_array_name,
                           required_shape = required_shape,
                           dtype = dtype)


def _cached_part_array_ref_const_notnone(collection, part, const_value, cached_array_name):
    assert not collection.points_for_part(part), 'constant arrays not supported for points properties'
    assert collection.support is not None
    shape = collection.supporting_shape(indexable_element = collection.indexable_for_part(part),
                                        direction = _part_direction(collection, part))
    assert shape is not None
    a = np.full(shape, const_value, dtype = float if collection.continuous_for_part(part) else int)
    setattr(collection, cached_array_name, a)


def _normalized_part_array_get_minmax(collection, trust_min_max, part, p_array, masked):
    min_value = max_value = None
    if trust_min_max:
        min_value = collection.minimum_value_for_part(part)
        max_value = collection.maximum_value_for_part(part)
    if min_value is None or max_value is None:
        min_value = np.nanmin(p_array)
        if masked and min_value is ma.masked:
            min_value = None
        max_value = np.nanmax(p_array)
        if masked and max_value is ma.masked:
            max_value = None
        collection.override_min_max(part, min_value, max_value)  # NB: this does not modify xml
    return min_value, max_value


def _realizations_array_ref_fill_missing(collection, r_extent, dtype, a):
    for part in collection.parts():
        realization = collection.realization_for_part(part)
        assert realization is not None and 0 <= realization < r_extent, 'realization missing (or out of range?)'
        pa = collection.cached_part_array_ref(part, dtype = dtype)
        a[realization] = pa
        collection.uncache_part_array(part)
    return a


def _realizations_array_ref_not_fill_missing(collection, r_list, dtype, a):
    for index in range(len(r_list)):
        realization = r_list[index]
        part = collection.singleton(realization = realization)
        pa = collection.cached_part_array_ref(part, dtype = dtype)
        a[index] = pa
        collection.uncache_part_array(part)
    return a


def _realizations_array_ref_get_shape_list(collection, indexable_element, r_extent):
    shape_list = collection.supporting_shape(indexable_element = indexable_element)
    shape_list.insert(0, r_extent)
    if collection.points_for_part(collection.parts()[0]):
        shape_list.append(3)
    return shape_list


def _realizations_array_ref_initial_checks(collection):
    assert collection.support is not None,  \
        'attempt to build realizations array for property collection without supporting representation'
    assert collection.number_of_parts() > 0,  \
        'attempt to build realizations array for empty property collection'
    assert collection.has_single_property_kind(),  \
        'attempt to build realizations array for collection with multiple property kinds'
    assert collection.has_single_indexable_element(),  \
        'attempt to build realizations array for collection containing a variety of indexable elements'
    assert collection.has_single_uom(),  \
        'attempt to build realizations array for collection containing multiple units of measure'
    r_list = collection.realization_list(sort_list = True)
    assert collection.number_of_parts() == len(r_list),  \
        'collection covers more than realizations of a single property'
    continuous = collection.all_continuous()
    if not continuous:
        assert collection.all_discrete(), 'mixture of continuous and discrete properties in collection'
    return r_list, continuous


def _time_array_ref_initial_checks(collection):
    assert collection.support is not None,  \
        'attempt to build time series array for property collection without supporting representation'
    assert collection.number_of_parts() > 0,  \
        'attempt to build time series array for empty property collection'
    assert collection.has_single_property_kind(),  \
        'attempt to build time series array for collection with multiple property kinds'
    assert collection.has_single_indexable_element(),  \
        'attempt to build time series array for collection containing a variety of indexable elements'
    assert collection.has_single_uom(),  \
        'attempt to build time series array for collection containing multiple units of measure'

    ti_list = collection.time_index_list(sort_list = True)
    assert collection.number_of_parts() == len(ti_list),  \
        'collection covers more than time indices of a single property'

    continuous = collection.all_continuous()
    if not continuous:
        assert collection.all_discrete(), 'mixture of continuous and discrete properties in collection'
    return ti_list, continuous


def _time_array_ref_fill_missing(collection, ti_extent, dtype, a):
    for part in collection.parts():
        time_index = collection.time_index_for_part(part)
        assert time_index is not None and 0 <= time_index < ti_extent, 'time index missing (or out of range?)'
        pa = collection.cached_part_array_ref(part, dtype = dtype)
        a[time_index] = pa
        collection.uncache_part_array(part)
    return a


def _time_array_ref_not_fill_missing(collection, ti_list, dtype, a):
    for index in range(len(ti_list)):
        time_index = ti_list[index]
        part = collection.singleton(time_index = time_index)
        pa = collection.cached_part_array_ref(part, dtype = dtype)
        a[index] = pa
        collection.uncache_part_array(part)
    return a


def _facet_array_ref_checks(collection):
    assert collection.support is not None,  \
        'attempt to build facets array for property collection without supporting representation'
    assert collection.number_of_parts() > 0,  \
        'attempt to build facets array for empty property collection'
    assert collection.has_single_property_kind(),  \
        'attempt to build facets array for collection containing multiple property kinds'
    assert collection.has_single_indexable_element(),  \
        'attempt to build facets array for collection containing a variety of indexable elements'
    assert collection.has_single_uom(),  \
        'attempt to build facets array for collection containing multiple units of measure'



"""Time duration"""

import logging

log = logging.getLogger(__name__)

import datetime as dt
import resqpy.time_series as rqts


class TimeDuration:
    """A thin wrapper around python's datetime timedelta objects (not a RESQML class)."""

    def __init__(self,
                 days = None,
                 hours = None,
                 minutes = None,
                 seconds = None,
                 earlier_timestamp = None,
                 later_timestamp = None):
        """Create a TimeDuration object either from days and seconds or from a pair of timestamps."""

        # for negative durations, earlier_timestamp should be later than later_timestamp
        # or days, hours etc. should typically all be non-positive
        # ie. days = -1, hours = -12 will be a negative one and a half day duration
        # whilst days = -1, hours = 12 will be a negative half day duration
        self.duration = None
        if earlier_timestamp is not None and later_timestamp is not None:
            rqts.check_timestamp(earlier_timestamp)
            rqts.check_timestamp(later_timestamp)
            if earlier_timestamp.endswith('Z'):
                earlier_timestamp = earlier_timestamp[:-1]  # Trailing Z is not part of iso format
            if later_timestamp.endswith('Z'):
                later_timestamp = later_timestamp[:-1]
            dt_earlier = dt.datetime.fromisoformat(earlier_timestamp)
            dt_later = dt.datetime.fromisoformat(later_timestamp)
            self.duration = dt_later - dt_earlier
        else:
            if days is None:
                days = 0
            if hours is None:
                hours = 0
            if minutes is None:
                minutes = 0
            if seconds is None:
                seconds = 0
            self.duration = dt.timedelta(days = days, hours = hours, minutes = minutes, seconds = seconds)

    def timestamp_after_duration(self, earlier_timestamp):
        """Create a new timestamp from this duration and an earlier timestamp."""

        if earlier_timestamp.endswith('Z'):
            earlier_timestamp = earlier_timestamp[:-1]
        rqts.check_timestamp(earlier_timestamp)
        dt_earlier = dt.datetime.fromisoformat(earlier_timestamp)
        dt_result = dt_earlier + self.duration
        return dt_result.isoformat() + 'Z'

    def timestamp_before_duration(self, later_timestamp):
        """Create a new timestamp from this duration and a later timestamp."""

        if later_timestamp.endswith('Z'):
            later_timestamp = later_timestamp[:-1]
        rqts.check_timestamp(later_timestamp)
        dt_later = dt.datetime.fromisoformat(later_timestamp)
        dt_result = dt_later - self.duration
        return dt_result.isoformat() + 'Z'



"""Class for wellintervalproperty, for resqml wellbore frame of blocked wellbore supports."""

import logging

log = logging.getLogger(__name__)

import resqpy.property
import resqpy.property.property_collection as rqp_pc


class WellIntervalProperty:
    """Thin wrapper class around interval properties for a Wellbore Frame or Blocked Wellbore.

    ie, interval or cell well logs.
    """

    def __init__(self, collection, part):
        """Create an interval log or blocked well log from a part name."""

        self.collection: rqp_pc.PropertyCollection = collection
        self.model = collection.model
        self.part = part

        indexable = self.collection.indexable_for_part(part)
        assert indexable in ['cells', 'intervals'], 'expected cells or intervals as indexable element'

        self.name = self.model.citation_title_for_part(part)
        self.uom = self.collection.uom_for_part(part)

    def values(self):
        """Return interval log or blocked well log as numpy array."""

        return self.collection.cached_part_array_ref(self.part)



"""Multiprocessing wrapper functions for the well/BlockedWell class."""

import logging

log = logging.getLogger(__name__)

import uuid
from typing import Tuple, Union, List
from pathlib import Path
from uuid import UUID

import resqpy.model as rq
import resqpy.grid as grr
import resqpy.well as rqw
import resqpy.multi_processing as rqmp


def blocked_well_from_trajectory_wrapper(
    index: int,
    parent_tmp_dir: str,
    grid_epc: str,
    grid_uuid: Union[UUID, str],
    trajectory_epc: str,
    trajectory_uuids: List[Union[UUID, str]],
) -> Tuple[int, bool, str, List[Union[UUID, str]]]:
    """Multiprocessing wrapper function of the BlockedWell initialisation from a Trajectory.

    arguments:
        index (int): the index of the function call from the multiprocessing function
        parent_tmp_dir (str): the parent temporary directory path from the multiprocessing function
        grid_epc (str): epc file path where the grid is saved
        grid_uuid (UUID or str): UUID (universally unique identifier) of the grid object
        trajectory_epc (str): epc file path where the trajectories are saved
        trajectory_uuids (list of UUID or str): a list of the trajectory uuids used to create each
            Trajectory object

    returns:
        Tuple containing:
        - index (int): the index passed to the function;
        - success (bool): True if all the BlockedWell objects could be created, False otherwise;
        - epc_file (str): the epc file path where the objects are stored;
        - uuid_list (List[UUID/str]): list of UUIDs of relevant objects

    note:
        used this wrapper when calling the multiprocessing function to initialise blocked wells from
        trajectories; it will create a new model that is saved in a temporary epc file
        and returns the required values, which are used in the multiprocessing function to
        recombine all the objects into a single epc file
    """
    uuid_list = []
    tmp_dir = Path(parent_tmp_dir) / f"{uuid.uuid4()}"
    tmp_dir.mkdir(parents = True, exist_ok = True)
    epc_file = str(tmp_dir / "wrapper.epc")
    model = rq.new_model(epc_file = epc_file, quiet = True)

    trajectory_model = rq.Model(trajectory_epc, quiet = True)
    grid_model = rq.Model(grid_epc)
    model.copy_uuid_from_other_model(grid_model, uuid = grid_uuid)

    grid = grr.any_grid(grid_model, uuid = grid_uuid)

    success = True
    for trajectory_uuid in trajectory_uuids:
        model.copy_uuid_from_other_model(trajectory_model, uuid = trajectory_uuid)
        trajectory = rqw.Trajectory(model, trajectory_uuid)
        blocked_well = rqw.BlockedWell(model, grid = grid, trajectory = trajectory)
        if blocked_well is None or blocked_well.cell_count is None or blocked_well.node_count is None:
            success = False
            continue
        blocked_well.write_hdf5()
        blocked_well.create_xml()
        uuid_list.append(blocked_well.uuid)

    model.store_epc(quiet = True)

    return index, success, epc_file, uuid_list


def blocked_well_from_trajectory_batch(grid_epc: str,
                                       grid_uuid: Union[UUID, str],
                                       trajectory_epc: str,
                                       trajectory_uuids: List[Union[UUID, str]],
                                       recombined_epc: str,
                                       cluster,
                                       n_workers: int,
                                       require_success: bool = False,
                                       tmp_dir_path: Union[Path, str] = '.') -> List[bool]:
    """Creates BlockedWell objects from a common grid and a list of trajectories' uuids, in parallel.

    arguments:
        grid_epc (str): epc file path where the grid is saved
        grid_uuid (UUID or str): UUID (universally unique identifier) of the grid object
        trajectory_epc (str): epc file path where the trajectories are saved
        trajectory_uuids (list of UUID or str): a list of the trajectory uuids used to create each
            Trajectory object
        recombined_epc (Path or str): A pathlib Path or path string, where the combined epc will be saved
        cluster (LocalCluster/JobQueueCluster): a LocalCluster is a Dask cluster on a
            local machine; if using a job queing system, a JobQueueCluster can be used
            such as an SGECluster, SLURMCluster, PBSCluster, LSFCluster etc
        n_workers (int): the number of workers on the cluster
        require_success (bool, default False): if True an exception is raised if any failures
        tmp_dir_path (str or Path, default '.'): the directory within which temporary directories will reside

    returns:
        success_list (list of bool): A boolean list of successful function calls

    notes:
        the returned success list contains one value per batch, set True if all blocked wells
        were successfully created in the batch, False if one or more failed in the batch
    """
    n_uuids = len(trajectory_uuids)
    trajectory_uuids_list = [
        trajectory_uuids[i * n_uuids // n_workers:(i + 1) * n_uuids // n_workers] for i in range(n_workers)
    ]

    kwargs_list = []
    for trajectory_uuids in trajectory_uuids_list:
        d = {
            "grid_epc": grid_epc,
            "grid_uuid": grid_uuid,
            "trajectory_epc": trajectory_epc,
            "trajectory_uuids": trajectory_uuids,
        }
        kwargs_list.append(d)

    success_list = rqmp.function_multiprocessing(blocked_well_from_trajectory_wrapper,
                                                 kwargs_list,
                                                 recombined_epc,
                                                 cluster,
                                                 require_success = require_success,
                                                 tmp_dir_path = tmp_dir_path)

    return success_list



"""High level extract box function."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.crs as rqc
import resqpy.derived_model
import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.box_utilities as bx
import resqpy.olio.fine_coarse as fc
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp

import resqpy.derived_model._common as rqdm_c


def extract_box(epc_file = None,
                source_grid = None,
                box = None,
                box_inactive = None,
                inherit_properties = False,
                inherit_realization = None,
                inherit_all_realizations = False,
                set_parent_window = None,
                new_grid_title = None,
                new_epc_file = None):
    """Extends an existing model with a new grid extracted as a logical IJK box from the source grid.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): if None, the epc_file is loaded and it should contain one ijk grid object
          (or one 'ROOT' grid) which is used as the source grid
       box (numpy int array of shape (2, 3)): the minimum and maximum kji0 indices in the source grid (zero based) to include
          in the extracted grid; note that cells with index equal to maximum value are included (unlike with python ranges)
       box_inactive (numpy bool array, optional): if present, shape must match box and values will be or'ed in with the
          inactive mask inherited from the source grid; if None, inactive mask will be as inherited from source grid
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid, with values taken from the specified box
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       set_parent_window (boolean, optional): if True, the extracted grid has its parent window attribute set; if False,
          the parent window is not set; if None, the default will be True if new_epc_file is None or False otherwise
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the extracted grid (& crs)

    returns:
       new grid object with extent as implied by the box argument

    note:
       the epc file and associated hdf5 file are appended to (extended) with the new grid, unless a new_epc_file is specified,
       in which case the grid and inherited properties are written there instead
    """

    log.debug('extracting grid for box')

    assert epc_file or new_epc_file, 'epc file name not specified'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    if set_parent_window is None:
        set_parent_window = (new_epc_file is None)
    model, source_grid = rqdm_c._establish_model_and_source_grid(epc_file, source_grid)
    assert source_grid.grid_representation in ['IjkGrid', 'IjkBlockGrid']
    assert model is not None
    assert box is not None and box.shape == (2, 3)
    assert np.all(box[1, :] >= box[0, :]) and np.all(box[0, :] >= 0) and np.all(box[1, :] < source_grid.extent_kji)

    if source_grid.grid_representation == 'IjkBlockGrid':
        source_grid.make_regular_points_cached()

    box_str = bx.string_iijjkk1_for_box_kji0(box)

    # create a new, empty grid object
    grid = grr.Grid(model)

    # inherit attributes from source grid
    grid.grid_representation = 'IjkGrid'
    grid.extent_kji = box[1, :] - box[0, :] + 1
    if box_inactive is not None:
        assert box_inactive.shape == tuple(grid.extent_kji)
    grid.nk, grid.nj, grid.ni = grid.extent_kji[0], grid.extent_kji[1], grid.extent_kji[2]
    grid.k_direction_is_down = source_grid.k_direction_is_down
    grid.grid_is_right_handed = source_grid.grid_is_right_handed
    grid.pillar_shape = source_grid.pillar_shape
    grid.has_split_coordinate_lines = source_grid.has_split_coordinate_lines
    # inherit the coordinate reference system used by the grid geometry
    grid.crs_uuid = source_grid.crs_uuid
    if source_grid.model is not model:
        model.duplicate_node(source_grid.model.root_for_uuid(grid.crs_uuid), add_as_part = True)
    grid.crs = rqc.Crs(model, uuid = grid.crs_uuid)

    # inherit k_gaps for selected layer range
    _inherit_k_gaps(source_grid, grid, box)

    # extract inactive cell mask
    _extract_inactive_cell_mask(source_grid, grid, box_inactive, box)

    # extract the grid geometry
    source_grid.cache_all_geometry_arrays()

    # determine cell geometry is defined
    if hasattr(source_grid, 'array_cell_geometry_is_defined'):
        grid.array_cell_geometry_is_defined = _array_box(source_grid.array_cell_geometry_is_defined, box)
        grid.geometry_defined_for_all_cells_cached = np.all(grid.array_cell_geometry_is_defined)
    else:
        grid.geometry_defined_for_all_cells_cached = source_grid.geometry_defined_for_all_cells_cached

    # copy info for pillar geometry is defined
    grid.geometry_defined_for_all_pillars_cached = source_grid.geometry_defined_for_all_pillars_cached
    if hasattr(source_grid, 'array_pillar_geometry_is_defined'):
        grid.array_pillar_geometry_is_defined = _array_box(source_grid.array_pillar_geometry_is_defined, box)
        grid.geometry_defined_for_all_pillars_cached = np.all(grid.array_pillar_geometry_is_defined)

    # get reference to points for source grid geometry
    source_points = source_grid.points_ref()

    pillar_box = box.copy()
    if source_grid.k_gaps:
        pillar_box[:, 0] = source_grid.k_raw_index_array[pillar_box[:, 0]]
    pillar_box[1, :] += 1  # pillar points have extent one greater than cells, in each axis

    if not source_grid.has_split_coordinate_lines:
        log.debug('no split pillars in source grid')
        grid.points_cached = _array_box(source_points, pillar_box)  # should work, ie. preserve xyz axis
    else:
        _process_split_pillars(source_grid, grid, box, pillar_box)

    if set_parent_window:
        fine_coarse = fc.FineCoarse(grid.extent_kji, grid.extent_kji, within_coarse_box = box)
        fine_coarse.set_all_ratios_constant()
        grid.set_parent(source_grid.uuid, True, fine_coarse)

    collection = _inherit_collection(source_grid, grid, inherit_properties, box, inherit_realization,
                                     inherit_all_realizations)

    if new_grid_title is None or len(new_grid_title) == 0:
        new_grid_title = 'local grid ' + box_str + ' extracted from ' + str(
            rqet.citation_title_for_node(source_grid.root))

    model.h5_release()
    if new_epc_file:
        rqdm_c._write_grid(new_epc_file,
                           grid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'w')
    else:
        ext_uuid, _ = model.h5_uuid_and_path_for_node(rqet.find_nested_tags(source_grid.root, ['Geometry', 'Points']),
                                                      'Coordinates')
        rqdm_c._write_grid(epc_file,
                           grid,
                           ext_uuid = ext_uuid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'a')

    return grid


def _array_box(a, box):
    return a[box[0, 0]:box[1, 0] + 1, box[0, 1]:box[1, 1] + 1, box[0, 2]:box[1, 2] + 1].copy()


def _local_col_index(extent, box, col):
    # return local equivalent natural column index for global column index, or None if outside box
    j, i = divmod(col, extent[2])
    j -= box[0, 1]
    i -= box[0, 2]
    if j < 0 or i < 0 or j > box[1, 1] - box[0, 1] or i > box[1, 2] - box[0, 2]:
        return None
    return j * (box[1, 2] - box[0, 2] + 1) + i


def _local_pillar_index(extent, box, p):
    # return local equivalent natural pillar index for global pillar index, or None if outside box
    p_j, p_i = divmod(p, extent[2] + 1)
    p_j -= box[0, 1]
    p_i -= box[0, 2]
    if p_j < 0 or p_i < 0 or p_j > box[1, 1] - box[0, 1] + 1 or p_i > box[1, 2] - box[0, 2] + 1:
        return None
    return p_j * (box[1, 2] - box[0, 2] + 2) + p_i


def _cols_for_pillar(extent, p):
    # return 4 naturalized column indices for columns surrounding natural pillar index; -1 where beyond edge of ij space
    cols = np.zeros((4,), dtype = int) - 1
    p_j, p_i = divmod(p, extent[2] + 1)
    if p_j > 0 and p_i > 0:
        cols[0] = (p_j - 1) * extent[2] + p_i - 1
    if p_j > 0 and p_i < extent[2]:
        cols[1] = (p_j - 1) * extent[2] + p_i
    if p_j < extent[1] and p_i > 0:
        cols[2] = p_j * extent[2] + p_i - 1
    if p_j < extent[1] and p_i < extent[2]:
        cols[3] = p_j * extent[2] + p_i
    return cols


def _inherit_k_gaps(source_grid, grid, box):
    if source_grid.k_gaps and box[1, 0] > box[0, 0]:
        k_gaps = np.count_nonzero(source_grid.k_gap_after_array[box[0, 0]:box[1, 0]])
        if k_gaps > 0:
            grid.k_gaps = k_gaps
            grid.k_gap_after_array = source_grid.k_gap_after_array[box[0, 0]:box[1, 0]].copy()
            grid.k_raw_index_array = np.empty(grid.nk, dtype = int)
            k_offset = 0
            for k in range(grid.nk):
                grid.k_raw_index_array[k] = k + k_offset
                if k < grid.nk - 1 and grid.k_gap_after_array[k]:
                    k_offset += 1
            assert k_offset == k_gaps


def _extract_inactive_cell_mask(source_grid, grid, box_inactive, box):
    if source_grid.inactive is None:
        if box_inactive is None:
            log.debug('setting inactive mask to None')
            grid.inactive = None
        else:
            log.debug('setting inactive mask to that passed as argument')
            grid.inactive = box_inactive.copy()
    else:
        if box_inactive is None:
            log.debug('extrating inactive mask')
            grid.inactive = _array_box(source_grid.inactive, box)
        else:
            log.debug('setting inactive mask to merge of source grid extraction and mask passed as argument')
            grid.inactive = np.logical_or(_array_box(source_grid.inactive, box), box_inactive)


def _inherit_collection(source_grid, grid, inherit_properties, box, inherit_realization, inherit_all_realizations):
    collection = None
    if inherit_properties:
        source_collection = source_grid.extract_property_collection()
        if source_collection is not None:
            # do not inherit the inactive property array by this mechanism
            active_collection = rqp.selective_version_of_collection(source_collection, property_kind = 'active')
            source_collection.remove_parts_list_from_dict(active_collection.parts())
            inactive_collection = rqp.selective_version_of_collection(
                source_collection,
                property_kind = 'code',  # for backward compatibility
                facet_type = 'what',
                facet = 'inactive')
            source_collection.remove_parts_list_from_dict(inactive_collection.parts())
            collection = rqp.GridPropertyCollection()
            collection.set_grid(grid)
            collection.extend_imported_list_copying_properties_from_other_grid_collection(
                source_collection,
                box = box,
                realization = inherit_realization,
                copy_all_realizations = inherit_all_realizations)
    return collection


def _process_split_pillars(source_grid, grid, box, pillar_box):
    source_points = source_grid.points_ref()
    source_base_pillar_count = (source_grid.nj + 1) * (source_grid.ni + 1)
    log.debug('number of base pillars in source grid: ' + str(source_base_pillar_count))
    log.debug('number of extra pillars in source grid: ' + str(len(source_grid.split_pillar_indices_cached)))
    base_points = _array_box(
        source_points[:, :source_base_pillar_count, :].reshape(
            (source_grid.nk_plus_k_gaps + 1, source_grid.nj + 1, source_grid.ni + 1, 3)),
        pillar_box).reshape(grid.nk_plus_k_gaps + 1, (grid.nj + 1) * (grid.ni + 1), 3)
    extra_points = np.zeros(
        (pillar_box[1, 0] - pillar_box[0, 0] + 1, source_points.shape[1] - source_base_pillar_count, 3))
    spi_array = np.zeros(len(source_grid.split_pillar_indices_cached), dtype = int)
    local_cols_array = np.zeros(len(source_grid.cols_for_split_pillars), dtype = int)
    local_cols_cl = np.zeros(len(source_grid.split_pillar_indices_cached), dtype = int)
    local_index = 0
    for index in range(len(source_grid.split_pillar_indices_cached)):
        source_pi = source_grid.split_pillar_indices_cached[index]
        local_pi = _local_pillar_index(source_grid.extent_kji, box, source_pi)
        if local_pi is None:
            continue
        cols = _cols_for_pillar(source_grid.extent_kji, source_pi)
        local_cols = _cols_for_pillar(grid.extent_kji, local_pi)
        if index == 0:
            start = 0
        else:
            start = source_grid.cols_for_split_pillars_cl[index - 1]
        finish = source_grid.cols_for_split_pillars_cl[index]
        source_pis = np.zeros((4,), dtype = int)
        for c_i in range(4):
            if local_cols[c_i] < 0:
                source_pis[c_i] = -1
                continue
            if cols[c_i] in source_grid.cols_for_split_pillars[start:finish]:
                source_pis[c_i] = source_base_pillar_count + index
            else:
                source_pis[c_i] = source_pi
        unique_source_pis = np.unique(source_pis)
        unique_count = len(unique_source_pis)
        unique_index = 0
        if unique_source_pis[0] == -1:
            unique_index = 1
            unique_count -= 1
        if unique_count <= 0:
            continue
        base_points[:, local_pi, :] = source_points[pillar_box[0, 0]:pillar_box[1, 0] + 1,
                                                    unique_source_pis[unique_index], :]
        unique_index += 1
        unique_count -= 1
        if unique_count <= 0:
            continue
        while unique_count > 0:
            source_pi = unique_source_pis[unique_index]
            extra_points[:, local_index, :] = source_points[pillar_box[0, 0]:pillar_box[1, 0] + 1, source_pi, :]
            spi_array[local_index] = local_pi
            if local_index == 0:
                lc_index = 0
            else:
                lc_index = local_cols_cl[local_index - 1]
            for c_i in range(4):
                if source_pis[c_i] == source_pi and local_cols[c_i] >= 0:
                    local_cols_array[lc_index] = local_cols[c_i]
                    lc_index += 1
            local_cols_cl[local_index] = lc_index
            local_index += 1
            unique_index += 1
            unique_count -= 1
    if local_index == 0:  # there are no split pillars in the box
        log.debug('box does not inherit any split pillars')
        grid.points_cached = base_points.reshape(grid.nk_plus_k_gaps + 1, grid.nj + 1, grid.ni + 1, 3)
        grid.has_split_coordinate_lines = False
    else:
        log.debug('number of extra pillars in box: ' + str(local_index))
        grid.points_cached = np.concatenate((base_points, extra_points[:, :local_index, :]), axis = 1)
        grid.split_pillar_indices_cached = spi_array[:local_index].copy()
        grid.cols_for_split_pillars = local_cols_array[:local_cols_cl[local_index - 1]].copy()
        grid.cols_for_split_pillars_cl = local_cols_cl[:local_index].copy()
        grid.split_pillars_count = local_index



"""Auxilliary functions for working with time series."""

# At present, no time zone information is handled

# This module differentiates between 'human timeframe' and 'geologic timeframe' time series, though RESQML does
#  not make such a distinction explicitly; the code here uses presence of YearOffset xml data to imply a time
# series on a geological timeframe

import logging

log = logging.getLogger(__name__)

import resqpy.time_series as rqts
import resqpy.olio.xml_et as rqet


def selected_time_series(full_series, indices_list, title = None):
    """Returns a new TimeSeries or GeologicTimeSeries object with timestamps selected from the full series by a list of indices."""

    from resqpy.time_series._geologic_time_series import GeologicTimeSeries
    from resqpy.time_series._time_series import TimeSeries

    if isinstance(full_series, TimeSeries):
        selected_ts = TimeSeries(full_series.model, title = title)
    else:
        assert isinstance(full_series, GeologicTimeSeries)
        selected_ts = GeologicTimeSeries(full_series.model, title = title)
    selected_ts.timestamps = [full_series.timestamps[i] for i in sorted(indices_list)]
    return selected_ts


def simplified_timestamp(timestamp):
    """Return a more readable version of the timestamp."""

    if timestamp is None:
        return None
    if isinstance(timestamp, int):
        return geologic_time_str(int)
    timestamp = cleaned_timestamp(timestamp)
    if len(timestamp) > 10:
        timestamp = timestamp[:10] + ' ' + timestamp[11:]
    return timestamp[:19]


def cleaned_timestamp(timestamp):
    """Return a cleaned version of the timestamp."""

    if timestamp is None:
        return None
    if isinstance(timestamp, int):
        return geologic_time_str(int)
    timestamp = str(timestamp)
    rqts.check_timestamp(timestamp)
    if len(timestamp) < 19 or timestamp[11:19] == '00:00:00':
        return timestamp[:10]
    return timestamp[:10] + 'T' + timestamp[11:19] + 'Z'


def time_series_from_list(timestamp_list, parent_model = None, title = None):
    """Create a TimeSeries object from a list of timestamps (model and node set to None).

    note:
       timestamps in the list should be in the correct string format for human timeframe series,
       or large negative integers for geologic timeframe series
    """

    from resqpy.time_series._geologic_time_series import GeologicTimeSeries
    from resqpy.time_series._time_series import TimeSeries

    assert (len(timestamp_list) > 0)
    sorted_timestamps = sorted(timestamp_list)
    if isinstance(sorted_timestamps[0], int):
        sorted_timestamps = sorted([-t if t > 0 else t for t in timestamp_list])
        time_series = GeologicTimeSeries(parent_model = parent_model, title = title)
        time_series.timestamps = sorted_timestamps
    else:
        sorted_timestamps = sorted(timestamp_list)
        time_series = TimeSeries(parent_model = parent_model,
                                 first_timestamp = cleaned_timestamp(sorted_timestamps[0]),
                                 title = title)
        for raw_timestamp in sorted_timestamps[1:]:
            timestamp = cleaned_timestamp(raw_timestamp)
            time_series.add_timestamp(timestamp)
    return time_series


def merge_timeseries_from_uuid(model, timeseries_uuid_iter):
    """Create a TimeSeries object from an iterable object of existing timeseries UUIDs of timeseries.

    iterable can be a list, array, or iterable generator (model must be provided). The new timeseries is sorted in
    ascending order. Returns the new time series, the new time series uuid, and the list of timeseries objects used to
    generate the list
    """

    reverse = False

    alltimestamps = set({})
    timeserieslist = []
    timeframe = None
    for timeseries_uuid in timeseries_uuid_iter:
        timeseriesroot = model.root(uuid = timeseries_uuid)
        assert (rqet.node_type(timeseriesroot) == 'obj_TimeSeries')

        singlets = any_time_series(model, uuid = timeseries_uuid)
        if timeframe is None:
            timeframe = singlets.timeframe
        else:
            assert timeframe == singlets.timeframe, 'attempt to merge human and geologic timeframe time series'

        timeserieslist.append(singlets)
        # alltimestamps.update( set(singlets.timestamps) )
        alltimestamps.update(set(singlets.datetimes()))

    sortedtimestamps = sorted(list(alltimestamps), reverse = reverse)

    new_time_series = time_series_from_list(sortedtimestamps, parent_model = model)
    new_time_series_uuid = new_time_series.uuid
    return new_time_series, new_time_series_uuid, timeserieslist


def geologic_time_str(years):
    """Returns a string representing a geological time for a large int representing number of years before present."""

    assert isinstance(years, int)
    years = abs(years)  # positive and negative values are both interpreted as the same: years before present
    if years < 10000000 and years % 1000000:
        stamp = f'{float(-years) / 1.0e6:.3f} Ma'
    else:
        stamp = f'{-years // 1000000} Ma'
    return stamp


def timeframe_for_time_series_uuid(model, uuid):
    """Returns string 'human' or 'geologic' indicating timeframe of the RESQML time series with a given uuid."""

    assert model is not None
    t = model.type_of_uuid(uuid = uuid, strip_obj = True)
    assert t is not None, 'time series uuid {uuid} not present in model'
    assert t == 'TimeSeries'

    root = model.root(uuid = uuid)

    em = rqet.load_metadata_from_xml(root)
    if em is not None:
        timeframe = em.get('timeframe')
        if timeframe:
            return timeframe

    t_node = rqet.find_tag(root, 'Time')
    assert t_node is not None

    return 'human' if rqet.find_tag(t_node, 'YearOffset') is None else 'geologic'


def any_time_series(parent_model, uuid):
    """Returns a resqpy TimeSeries or GeologicTimeSeries object for an existing RESQML time series with a given uuid."""

    from resqpy.time_series._geologic_time_series import GeologicTimeSeries
    from resqpy.time_series._time_series import TimeSeries

    timeframe = timeframe_for_time_series_uuid(parent_model, uuid)
    if timeframe == 'human':
        return TimeSeries(parent_model, uuid = uuid)
    assert timeframe == 'geologic'
    return GeologicTimeSeries(parent_model, uuid = uuid)



"""Support for consolidation of datasets based on equivalence between parts."""

import logging

log = logging.getLogger(__name__)

import resqpy.crs as rqc
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
import resqpy.property as rqp
import resqpy.time_series as rqt

# the following list contains those RESQML classes with equivalence methods in their resqpy class
# it is lightly ordered with earlier classes having no dependence on classes later in the list

consolidatable_list = [
    'OrganizationFeature', 'GeobodyFeature', 'BoundaryFeature', 'FrontierFeature', 'GeologicUnitFeature',
    'FluidBoundaryFeature', 'RockFluidUnitFeature', 'TectonicBoundaryFeature', 'GeneticBoundaryFeature',
    'WellboreFeature', 'FaultInterpretation', 'EarthModelInterpretation', 'HorizonInterpretation',
    'GeobodyBoundaryInterpretation', 'GeobodyInterpretation', 'WellboreInterpretation', 'LocalDepth3dCrs',
    'LocalTime3dCrs', 'TimeSeries', 'StringTableLookup', 'PropertyKind'
]

# todo: add to this list as other classes gain an is_equivalent() method


class Consolidation:
    """Class supporting equivalence mapping of high level RESQML parts between models."""

    def __init__(self, resident_model):
        """Initialise a new Consolidation object prior to merging parts from another model.

        arguments:
           resident_model (model.Model): the model into which potentially equivalent parts will be merged

        returns:
           the new Consolidation object
        """

        self.model = resident_model
        log.debug(f'new consolidation for {self.model.epc_file} with {len(self.model.uuids())} uuids')
        self.map = {}  # dictionary mapping immigrant uuid int to primary uuid int
        self.stale = True

    def equivalent_uuid_for_part(self, part, immigrant_model = None, ignore_identical_part = False):
        """Returns uuid of an equivalent part in resident model, or None if no equivalent found."""
        uuid_int = self.equivalent_uuid_int_for_part(part,
                                                     immigrant_model = immigrant_model,
                                                     ignore_identical_part = ignore_identical_part)
        return bu.uuid_from_int(uuid_int)

    def equivalent_uuid_int_for_part(self, part, immigrant_model = None, ignore_identical_part = False):
        """Returns uuid.int of an equivalent part in resident model, or None if no equivalent found."""

        # log.debug('Looking for equivalent uuid for: ' + str(part))
        if not part:
            return None
        if immigrant_model is None:
            immigrant_model = self.model
        immigrant_uuid_int = rqet.uuid_in_part_name(part).int
        # log.debug('   immigrant uuid: ' + str(immigrant_uuid))
        if immigrant_uuid_int in self.map:
            # log.debug('   known to be equivalent to: ' + str(self.map[immigrant_uuid_int]))
            return self.map[immigrant_uuid_int]
        obj_type = immigrant_model.type_of_part(part, strip_obj = True)
        if obj_type is None or obj_type not in consolidatable_list:
            return None
        # log.debug('   object type is consolidatable')
        resident_uuids = self.model.uuids(obj_type = obj_type)
        if resident_uuids is None or len(resident_uuids) == 0:
            # log.debug('   no resident parts found of type: ' + str(obj_type))
            return None
        # log.debug(f'   {len(resident_uuids)} resident parts of same class')
        if not ignore_identical_part:
            for resident_uuid in resident_uuids:
                if resident_uuid.int == immigrant_uuid_int:
                    # log.debug('   uuid already resident: ' + str(resident_uuid))
                    return resident_uuid.int

        # log.debug('   preparing immigrant object')
        immigrant_uuid = bu.uuid_from_int(immigrant_uuid_int)
        if obj_type.endswith('Interpretation') or obj_type.endswith('Feature'):
            immigrant_obj = rqo.__dict__[obj_type](immigrant_model, uuid = immigrant_uuid)
        elif obj_type.endswith('Crs'):
            immigrant_obj = rqc.Crs(immigrant_model, uuid = immigrant_uuid)
        elif obj_type == 'TimeSeries':
            immigrant_obj = rqt.TimeSeries(immigrant_model, uuid = immigrant_uuid)
        elif obj_type == 'StringTableLookup':
            immigrant_obj = rqp.StringLookup(immigrant_model, uuid = immigrant_uuid)
        elif obj_type == 'PropertyKind':
            immigrant_obj = rqp.PropertyKind(immigrant_model, uuid = immigrant_uuid)
        else:
            raise Exception('code failure')
        assert immigrant_obj is not None
        for resident_uuid in resident_uuids:
            resident_uuid_int = resident_uuid.int
            # log.debug('   considering resident: ' + str(resident_uuid))
            if ignore_identical_part and bu.matching_uuids(resident_uuid, immigrant_uuid):
                continue
            if obj_type.endswith('Interpretation') or obj_type.endswith('Feature'):
                resident_obj = rqo.__dict__[obj_type](self.model, uuid = resident_uuid)
            elif obj_type.endswith('Crs'):
                resident_obj = rqc.Crs(self.model, uuid = resident_uuid)
            elif obj_type == 'TimeSeries':
                resident_obj = rqt.TimeSeries(self.model, uuid = resident_uuid)
            elif obj_type == 'StringTableLookup':
                resident_obj = rqp.StringLookup(self.model, uuid = resident_uuid)
            elif obj_type == 'PropertyKind':
                resident_obj = rqp.PropertyKind(self.model, uuid = resident_uuid)
            else:
                raise Exception('code failure')
            assert resident_obj is not None
            # log.debug('   comparing with: ' + str(resident_obj.uuid))
            if immigrant_obj == resident_obj:  # note: == operator overloaded with equivalence method for these classes
                while resident_uuid_int in self.map:
                    # log.debug('   following equivalence for: ' + str(resident_uuid))
                    resident_uuid_int = self.map[resident_uuid_int]
                self.map[immigrant_uuid_int] = resident_uuid_int
                # log.debug('   new equivalence found with: ' + str(resident_uuid))
                return resident_uuid_int
        return None

    def force_uuid_equivalence(self, immigrant_uuid, resident_uuid):
        """Forces immigrant object to be treated as equivalent to (same as) resident object, identified by uuids."""

        assert immigrant_uuid is not None and resident_uuid is not None
        if isinstance(immigrant_uuid, str):
            immigrant_uuid = bu.uuid_from_string(immigrant_uuid)
        if isinstance(resident_uuid, str):
            resident_uuid = bu.uuid_from_string(resident_uuid)
        if bu.matching_uuids(immigrant_uuid, resident_uuid):
            return
        assert immigrant_uuid not in self.map.values()

        self.force_uuid_int_equivalence(immigrant_uuid.int, resident_uuid.int)

    def force_uuid_int_equivalence(self, immigrant_uuid_int, resident_uuid_int):
        """Forces immigrant object to be treated as equivalent to (same as) resident object, identified by uuid ints."""

        assert immigrant_uuid_int is not None and resident_uuid_int is not None
        if immigrant_uuid_int == resident_uuid_int:
            return
        assert immigrant_uuid_int not in self.map.values()

        self.map[immigrant_uuid_int] = resident_uuid_int

    def force_part_equivalence(self, immigrant_part, resident_part):
        """Forces immigrant part to be treated as equivalent to resident part."""

        assert immigrant_part is not None and resident_part is not None
        if immigrant_part == resident_part:
            return
        self.force_uuid_int_equivalence(
            rqet.uuid_in_part_name(immigrant_part).int,
            rqet.uuid_in_part_name(resident_part).int)

    def check_map_integrity(self):
        """Raises assertion failure if map contains any potentially circular references."""

        for immigrant in self.map.keys():
            assert immigrant not in self.map.values()
        for resident in self.map.values():
            assert resident not in self.map.keys()


def _ordering(obj_type):
    if obj_type in consolidatable_list:
        return consolidatable_list.index(obj_type)
    seq = len(consolidatable_list)
    if obj_type.endswith('Interpretation'):
        seq += 1
    elif obj_type.endswith('Representation'):
        seq += 2
    elif obj_type.endswith('Property'):
        seq += 3
    return seq


def sort_parts_list(model, parts_list):
    """Returns a copy of the parts list sorted into the preferred order for consolidating."""

    pair_list = [(_ordering(model.type_of_part(part, strip_obj = True)), part) for part in parts_list]
    pair_list.sort()
    return [part for _, part in pair_list]


def sort_uuids_list(model, uuids_list):
    """Returns a copy of the uuids list (or uuid ints list) sorted into the preferred order for consolidating."""

    pair_list = [
        (_ordering(model.type_of_part(model.part_for_uuid(uuid), strip_obj = True)), uuid) for uuid in uuids_list
    ]
    pair_list.sort()
    return [uuid for _, uuid in pair_list]



"""MdDatum class."""

# Nexus is a registered trademark of the Halliburton Company
# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import warnings
import numpy as np

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns

valid_md_reference_list = [
    "ground level", "kelly bushing", "mean sea level", "derrick floor", "casing flange", "arbitrary point",
    "crown valve", "rotary bushing", "rotary table", "sea floor", "lowest astronomical tide", "mean higher high water",
    "mean high water", "mean lower low water", "mean low water", "mean tide level", "kickoff point"
]

# todo: could require/maintain DeviationSurvey mds in same units as md datum object's crs vertical units?


class MdDatum(BaseResqpy):
    """Class for RESQML measured depth datum."""

    resqml_type = 'MdDatum'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 crs_uuid = None,
                 location = None,
                 md_reference = 'mean sea level',
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Initialises a new MdDatum object.

        arguments:
           parent_model (model.Model object): the model which the new md datum belongs to
           uuid: If not None, load from existing object. Else, create new.
           crs_uuid (uuid.UUID): required if initialising from values
           location: (triple float): the x, y, z location of the new measured depth datum;
              ignored if uuid is not None
           md_reference (string): human readable resqml standard string indicating the real
              world nature of the datum, eg. 'kelly bushing'; the full list of options is
              available as the global variable valid_md_reference_list in this module;
              ignored if uuid is not None
           title (str, optional): the citation title to use for a new datum;
              ignored if uuid is not None
           originator (str, optional): the name of the person creating the datum, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the datum;
              ignored if uuid is not None

        returns:
           the newly instantiated measured depth datum object

        note:
           this function does not create an xml node for the md datum; call the create_xml() method afterwards
           if initialising from data other than an existing RESQML object
        """

        self.location = location
        self.md_reference = md_reference
        self.crs_uuid = crs_uuid

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is None and (location is not None or md_reference):
            assert location is not None and md_reference
            assert md_reference in valid_md_reference_list
            assert len(location) == 3

    def _load_from_xml(self):
        md_datum_root = self.root
        assert md_datum_root is not None
        location_node = rqet.find_tag(md_datum_root, 'Location')
        self.location = (rqet.find_tag_float(location_node,
                                             'Coordinate1'), rqet.find_tag_float(location_node, 'Coordinate2'),
                         rqet.find_tag_float(location_node, 'Coordinate3'))
        self.md_reference = rqet.node_text(rqet.find_tag(md_datum_root, 'MdReference')).strip().lower()
        assert self.md_reference in valid_md_reference_list
        self.crs_uuid = self.extract_crs_uuid()

    # todo: the following function is almost identical to one in the grid module: it should be made common and put in model.py

    def extract_crs_uuid(self):
        """Returns uuid for coordinate reference system, as stored in reference node of this md datum's xml tree."""

        if self.crs_uuid is not None:
            return self.crs_uuid
        crs_root = rqet.find_tag(self.root, 'LocalCrs')
        uuid_str = rqet.find_tag(crs_root, 'UUID').text
        self.crs_uuid = bu.uuid_from_string(uuid_str)
        return self.crs_uuid

    def create_part(self):
        """Creates xml for this md datum object and adds to parent model as a part; returns root node for part."""

        # note: deprecated, call create_xml() directly
        assert self.root is None
        assert self.location is not None
        self.create_xml(add_as_part = True)

    def create_xml(self, add_as_part = True, add_relationships = True, title = None, originator = None):
        """Creates xml for a measured depth datum element; crs node must already exist; optionally adds as part.

        arguments:
           add_as_part (boolean, default True): if True, the newly created xml node is added as a part
              in the model
           add_relationships (boolean, default True): if True, a relationship xml part is created relating the
              new md datum part to the crs
           title (string): used as the citation Title text for the new md datum node
           originator (string, optional): the name of the human being who created the md datum part;
              default is to use the login name

        returns:
           the newly created measured depth datum xml node
        """

        md_reference = self.md_reference.lower()
        assert md_reference in valid_md_reference_list, 'invalid measured depth reference: ' + md_reference

        if title:
            self.title = title
        if not self.title:
            self.title = 'measured depth datum'

        crs_uuid = self.crs_uuid
        assert crs_uuid is not None

        datum = super().create_xml(add_as_part = False, originator = originator)

        self.model.create_solitary_point3d('Location', datum, self.location)

        md_ref = rqet.SubElement(datum, ns['resqml2'] + 'MdReference')
        md_ref.set(ns['xsi'] + 'type', ns['resqml2'] + 'MdReference')
        md_ref.text = md_reference

        self.model.create_crs_reference(crs_uuid = crs_uuid, root = datum)

        if add_as_part:
            self.model.add_part('obj_MdDatum', self.uuid, datum)
            if add_relationships:
                crs_root = self.model.root_for_uuid(self.crs_uuid)
                self.model.create_reciprocal_relationship(datum, 'destinationObject', crs_root, 'sourceObject')

        return datum

    def is_equivalent(self, other):
        """Implements equals operator, comparing metadata items deemed significant."""

        if not isinstance(other, self.__class__):
            return False
        if self.md_reference != other.md_reference or not np.allclose(self.location, other.location):
            return False
        return bu.matching_uuids(self.crs_uuid, other.crs_uuid)
